# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import xml.etree.ElementTree as ETree
import urllib.parse
import urllib.request
from xl.dynamic import DynamicSource
from xl import providers

import logging

logger = logging.getLogger(__name__)

LFMS = None

# Last.fm API Key for Exaile
# if you reuse this code in a different application, please
# register your own key with last.fm
API_KEY = '3599c79a97fd61ce518b75922688bc38'


def enable(exaile):
    global LFMS
    LFMS = LastfmSource()
    providers.register("dynamic_playlists", LFMS)


def disable(exaile):
    global LFMS
    providers.unregister("dynamic_playlists", LFMS)
    LFMS = None


class LastfmSource(DynamicSource):
    name = 'lastfm'

    def __init__(self):
        DynamicSource.__init__(self)

    def get_results(self, artist):
        ar = urllib.parse.quote_plus(artist.encode('utf-8'))
        url = (
            'https://ws.audioscrobbler.com/2.0/?method=artist.getsimilar&artist=%s&api_key='
            + API_KEY
        )
        try:
            f = urllib.request.urlopen(url % ar).read()
        except IOError:
            logger.exception("Error retrieving results")
            return []

        retlist = []
        xml = ETree.fromstring(f)

        for e in xml.getiterator('artist'):
            retlist.append((float(e.find('match').text), e.find('name').text))

        return retlist



# Copyright (C) 2014 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk

from xl.nls import gettext as _, ngettext

from xlgui.guiutil import GtkTemplate
from xlgui.widgets import dialogs

from . import gt_common


@GtkTemplate('gt_mass.ui', relto=__file__)
class GtMassRename(Gtk.Window):
    __gtype_name__ = 'GTMassRename'

    (
        found_label,
        playlists,
        replace,
        replace_entry,
        search_entry,
        tracks_list,
    ) = GtkTemplate.Child.widgets(6)

    def __init__(self, exaile):
        Gtk.Window.__init__(self, transient_for=exaile.gui.main.window)
        self.init_template()

        self.exaile = exaile

        self.tracks_list.get_model().set_sort_column_id(1, Gtk.SortType.ASCENDING)

        # initialize playlist list
        model = self.playlists.get_model()

        for pl in exaile.smart_playlists.list_playlists():
            model.append((True, pl))

        for pl in exaile.playlists.list_playlists():
            model.append((False, pl))

        self.show_all()

    def reset(self):
        self.tracks_list.get_model().clear()
        self.replace.set_sensitive(False)

    @GtkTemplate.Callback
    def on_find_clicked(self, widget):
        self.search_str = self.search_entry.get_text().strip()
        self.tagname = gt_common.get_tagname()

        # freeze update
        model = self.tracks_list.get_model()
        self.tracks_list.freeze_child_notify()
        self.tracks_list.set_model(None)

        model.clear()

        idx = self.playlists.get_active()
        if idx != -1 and self.search_str != '':
            smart, name = self.playlists.get_model()[idx]
            if smart:
                pl = self.exaile.smart_playlists.get_playlist(name)
            else:
                pl = self.exaile.playlists.get_playlist(name)

            if hasattr(pl, 'get_playlist'):
                pl = pl.get_playlist(self.exaile.collection)

            for track in pl:
                groups = gt_common._get_track_groups(track, self.tagname)

                if self.search_str != '' and self.search_str not in groups:
                    continue

                name = ' - '.join(
                    [
                        track.get_tag_display('artist'),
                        track.get_tag_display('album'),
                        track.get_tag_display('title'),
                    ]
                )
                model.append((True, name, track))

        # unfreeze, draw it up
        self.tracks_list.set_model(model)
        self.tracks_list.thaw_child_notify()

        self.found_label.set_text(
            ngettext(
                '{amount} track found', '{amount} tracks found', len(model)
            ).format(amount=len(model))
        )

        self.replace.set_sensitive(len(model) != 0)

    @GtkTemplate.Callback
    def on_replace_clicked(self, widget):
        tracks = [row[2] for row in self.tracks_list.get_model() if row[1]]
        replace_str = self.replace_entry.get_text().strip()

        if replace_str:
            query = ngettext(
                "Replace '{old_tag}' with '{new_tag}' on {amount} track?",
                "Replace '{old_tag}' with '{new_tag}' on {amount} tracks?",
                len(tracks),
            ).format(
                old_tag=self.search_str,
                new_tag=replace_str,
                amount=len(tracks),
            )
        else:
            query = ngettext(
                "Delete '{tag}' from {amount} track?",
                "Delete '{tag}' from {amount} tracks?",
                len(tracks),
            ).format(tag=self.search_str, amount=len(tracks))

        if dialogs.yesno(self, query) != Gtk.ResponseType.YES:
            return

        for track in tracks:
            groups = gt_common._get_track_groups(track, self.tagname)

            if self.search_str != '':
                groups.discard(self.search_str)

            if replace_str != '':
                groups.add(replace_str)

            if not gt_common.set_track_groups(track, groups):
                return

        dialogs.info(self, "Tags successfully renamed!")
        self.reset()


def mass_rename(exaile):
    message = _(
        "You should rescan your collection before using mass tag "
        "rename to ensure that all tags are up to date. Rescan now?"
    )

    if dialogs.yesno(exaile.gui.main.window, message) == Gtk.ResponseType.YES:
        exaile.gui.on_rescan_collection()

    GtMassRename(exaile)



# moodbar - Replace Exaile's seekbar with a moodbar
# Copyright (C) 2015, 2018-2019, 2021  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Tuple

import cairo


class Painter:
    def paint(self, data: bytes) -> cairo.ImageSurface:
        """Paint moodbar to a Cairo surface.

        :param data: Moodbar data
        :return: Cairo surface containing the image to be drawn
        """
        raise NotImplementedError


class NormalPainter(Painter):
    def paint(self, data: bytes) -> cairo.ImageSurface:
        width = len(data) // 3
        surf = cairo.ImageSurface(cairo.FORMAT_RGB24, width, 1)
        arr = surf.get_data()
        for index in range(width):
            index4 = index * 4
            index3 = index * 3
            # Cairo RGB24 is BGRX
            r, g, b = data[index3 : index3 + 3]
            arr[index4 + 0] = b
            arr[index4 + 1] = g
            arr[index4 + 2] = r
        return surf


class WaveformPainter(Painter):
    """Paint a waveform-like representation of the moodbar"""

    def paint(self, data: bytes) -> cairo.ImageSurface:
        H = 100  # Number of pixels on one side of the waveform
        TOTAL_H = H * 2 + 1  # 1 px in the middle + H px above + H px below

        width = len(data) // 3
        surf = cairo.ImageSurface(cairo.FORMAT_RGB24, width, TOTAL_H)
        stride = surf.get_stride()
        arr = surf.get_data()
        for index in range(width):
            index3 = index * 3
            index4 = index * 4
            rgb = data[index3 : index3 + 3]
            level = int(round(self._scale_level(*rgb) * H))
            ystart = H - level
            yend = H + level
            rgb = self._scale_color(*rgb)
            for ic in range(3):
                # Cairo RGB24 is BGRX
                val = rgb[2 - ic]
                for i in range(
                    ystart * stride + index4 + ic,
                    yend * stride + index4 + ic + 1,
                    stride,
                ):
                    arr[i] = val
        return surf

    @staticmethod
    def _scale_color(r: int, g: int, b: int) -> Tuple[int, int, int]:
        """Modify a color so it looks nicer on the waveform moodbar."""

        import colorsys
        import math

        def _clamp(x, low, high):
            return max(low, min(x, high))

        h, s, v = colorsys.rgb_to_hsv(r / 255, g / 255, b / 255)

        # These numbers are pulled out of thin air
        MIN = 80 / 255
        v = v * (1 - MIN) + MIN
        v = math.log(v + 1, 2)

        rf, gf, bf = colorsys.hsv_to_rgb(h, s, v)

        def to_u8(c):
            return int(_clamp(c, 0, 1) * 255)

        return to_u8(rf), to_u8(gf), to_u8(bf)

    @staticmethod
    def _scale_level(r: int, g: int, b: int) -> float:
        """Calculate level based on moodbar color"""

        import math

        level = math.sqrt(r**2 + g**2 + b**2)
        level /= math.sqrt(255**2 * 3)

        # These numbers are pulled out of thin air
        MIN = 10 / 255
        level = level * (1 - MIN) + MIN
        level = math.log(level + 1, 2)

        return level



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

from gi.repository import Gtk, Gdk

from xl import common, providers
from xlgui.accelerators import Accelerator

# Fake accel group so that menuitems can trick GTK into
# showing accelerators in the menus.
FAKEACCELGROUP = Gtk.AccelGroup()


def simple_separator(name, after):
    def factory(menu, parent, context):
        item = Gtk.SeparatorMenuItem()
        return item

    item = MenuItem(name, factory, after=after)
    item._pos = 'last'
    return item


def _get_accel(
    callback: Union[Accelerator, Callable, None], display_name: Optional[str]
) -> Tuple[Optional[Accelerator], Optional[Callable], Optional[str]]:
    # utility function to get menu information from an xlgui.Accelerator
    accelerator: Optional[Accelerator] = None
    if isinstance(callback, Accelerator):
        accelerator = callback
        if display_name is None:
            display_name = accelerator.helptext
        callback = accelerator.callback
    return accelerator, callback, display_name


def simple_menu_item(
    name: str,
    after: List[str],
    display_name: Optional[str] = None,
    icon_name: Optional[str] = None,
    callback: Union[Accelerator, Callable, None] = None,
    callback_args: Iterable[Any] = [],
    submenu: Optional[Gtk.Menu] = None,
    condition_fn: Optional[Callable[..., bool]] = None,
    sensitive_cb: Optional[Callable[[], bool]] = None,
) -> 'MenuItem':
    """
    Factory function that should handle most cases for menus

    :param name: Internal name for the item. must be unique within the menu.
    :param after: List of ids which come before this item, this item will
            be placed after the lowest of these.
    :param display_name: Name as is to appear in the menu.
    :param icon_name: Name of the icon to display, or None for no icon.
    :param callback: The function to call when the menu item is activated OR
                     an xlgui accelerator object.
            signature: callback(widget, name, parent, context)
    :param submenu: The Gtk.Menu that is to be the submenu of this item
    :param condition_fn: A function to call when the menu is displayed. If
            the function returns False, the menu item is not shown
            signature: condition_fn(name, parent, context)
    :param sensitive_cb: A function that if it returns False, the menu item
                         will be disabled
    """
    accelerator, callback, display_name = _get_accel(callback, display_name)

    def factory(menu, parent, context) -> Optional[Gtk.MenuItem]:
        item: Optional[Gtk.MenuItem] = None

        if condition_fn is not None and not condition_fn(name, parent, context):
            return None

        if display_name is not None:
            if icon_name is not None:
                item = Gtk.ImageMenuItem.new_from_stock(display_name)
                image = Gtk.Image.new_from_icon_name(icon_name, size=Gtk.IconSize.MENU)
                item.set_image(image)
            else:
                item = Gtk.MenuItem.new_with_mnemonic(display_name)
        else:
            assert icon_name is not None
            item = Gtk.ImageMenuItem.new_from_stock(icon_name)

        if submenu is not None:
            item.set_submenu(submenu)

        if accelerator is not None:
            item.add_accelerator(
                'activate',
                FAKEACCELGROUP,
                accelerator.key,
                accelerator.mods,
                Gtk.AccelFlags.VISIBLE,
            )

        if callback is not None:
            item.connect('activate', callback, name, parent, context, *callback_args)

        if sensitive_cb is not None and not sensitive_cb():
            item.set_sensitive(False)

        return item

    return MenuItem(name, factory, after=after)


def check_menu_item(
    name: str, after, display_name: str, checked_func: Callable[..., bool], callback
) -> 'MenuItem':
    accelerator, callback, display_name = _get_accel(callback, display_name)

    def factory(menu, parent, context):
        item = Gtk.CheckMenuItem.new_with_mnemonic(display_name)
        active = checked_func(name, parent, context)
        item.set_active(active)
        if accelerator is not None:
            item.add_accelerator(
                'activate',
                FAKEACCELGROUP,
                accelerator.key,
                accelerator.mods,
                Gtk.AccelFlags.VISIBLE,
            )
        item.connect('activate', callback, name, parent, context)
        return item

    return MenuItem(name, factory, after=after)


def radio_menu_item(
    name: str, after, display_name: str, groupname, selected_func, callback
) -> 'RadioMenuItem':
    def factory(menu, parent, context):
        for index, item in enumerate(menu._items):
            if hasattr(item, 'groupname') and item.groupname == groupname:
                break
        else:
            index = None

        if index is not None:
            try:
                group_parent = menu.get_children()[index]
                if not isinstance(group_parent, Gtk.RadioMenuItem):
                    group_parent = None
            except IndexError:
                group_parent = None

        if group_parent:
            group = group_parent.get_group()
        else:
            group = None

        item = Gtk.RadioMenuItem.new_with_mnemonic(group, display_name)
        active = selected_func(name, parent, context)
        item.set_active(active)

        item.connect('activate', callback, name, parent, context)
        return item

    return RadioMenuItem(name, factory, after=after, groupname=groupname)


class MenuItem:
    __slots__ = ['name', 'after', '_factory', '_pos', '_provider_data']

    def __init__(
        self,
        name: str,
        factory: Optional[Callable[..., Optional[Gtk.MenuItem]]],
        after: List[str],
    ):
        self.name = name
        self.after = after
        self._factory = factory
        # Don't change this unless you have a REALLY good
        # reason to. after= is the 'supported'
        # method of ordering, this property is not
        # considered public api and may change
        # without warning.
        self._pos = 'normal'

    def factory(self, menu, parent, context) -> Optional[Gtk.MenuItem]:
        """
        The factory function is called when the menu is shown, and
        should return a menu item. If it returns None, the item is
        not shown.
        """
        return self._factory(menu, parent, context)

    def register(self, servicename, target=None):
        """
        Shortcut for providers.register(), allows registration
        for use with a ProviderMenu
        """
        self._provider_data = (servicename, target)
        return providers.register(servicename, self, target=target)

    def unregister(self):
        """
        Shortcut for providers.unregister()
        """
        servicename, target = self._provider_data
        return providers.unregister(servicename, self, target)

    def __repr__(self):
        return '<xlgui.widgets.MenuItem: %s>' % self.name


class RadioMenuItem(MenuItem):
    __slots__ = ['groupname']

    def __init__(self, name, factory, after, groupname):
        MenuItem.__init__(self, name, factory, after)
        self.groupname = groupname


class Menu(Gtk.Menu):
    """
    Generic flexible menu with reliable
    menu item order and context handling
    """

    def __init__(self, parent, context_func=None, inherit_context=False):
        """
        :param parent: the parent for this menu
        :param context_func: a function for context
            retrieval
        :param inherit_context: If a submenu, inherit context function from
                                parent menu
        """
        Gtk.Menu.__init__(self)
        self._parent = parent
        self._items: List[MenuItem] = []
        self.context_func = context_func
        self.connect('show', lambda *e: self.regenerate_menu())
        self.connect('hide', lambda *e: self.clear_menu())
        # Placeholder exists to make sure unity doesn't get confused (legacy?)
        self.placeholder = Gtk.MenuItem.new_with_mnemonic('')
        self._inherit_context = inherit_context

    def get_context(self):
        """
        Retrieves the menu context which
        can contain various data

        :returns: {'key1': 'value1', ...}
        :rtype: dictionary
        """
        if self._inherit_context:
            return self.get_parent_shell().get_context()
        elif self.context_func is None:
            return {}
        else:
            return self.context_func(self._parent)

    def add_item(self, item: MenuItem) -> None:
        """
        Adds a menu item and triggers reordering

        :param item: the menu item
        """
        self._items.append(item)
        self.reorder_items()

    def add_simple(self, label: str, callback, icon_name: Optional[str] = None) -> None:
        """
        Provide a simple mechanism to add menu items without a lot of hassle

        :param label: Label to display
        :param callback: Callback that will be called on click
        :param icon_name: GTK mostly ignores this, and it will go away

        .. note:: If you use this API, you should generally only use this
                  API to add items to that menu
        """
        self.add_item(
            simple_menu_item(
                '_i%s' % len(self._items),
                [],
                label,
                icon_name=icon_name,
                callback=callback,
            )
        )

    def remove_item(self, item: MenuItem) -> None:
        """
        Removes a menu item

        :param item: the menu item
        """
        self._items.remove(item)

    def clear_menu(self) -> None:
        """
        Removes all menu items and submenus to prevent
        references sticking around due to saved contexts
        """
        self.append(self.placeholder)
        children = self.get_children()
        for c in children:
            if c is self.placeholder:
                continue
            c.set_submenu(None)
            self.remove(c)

    def reorder_items(self) -> None:
        """
        Reorders all menu items
        """
        pmap = {'first': 0, 'normal': 1, 'last': 2}
        items = [
            common.PosetItem(i.name, i.after, pmap[i._pos], value=i)
            for i in self._items
        ]
        items = common.order_poset(items)
        self._items = [i.value for i in items]

    def regenerate_menu(self) -> None:
        """
        Regenerates the menu by retrieving
        the context and calling the factory
        method of all menu items
        """
        context = self.get_context()
        if self.placeholder in self.get_children():
            self.remove(self.placeholder)
        for item in self._items:
            subitem = item.factory(self, self._parent, context)
            if subitem is not None:
                self.append(subitem)
        self.show_all()

    def popup(self, *args) -> None:
        """
        Pops out the menu (Only if
        there are items to show)
        """
        if len(self._items) > 0:
            if len(args) == 1:
                event = args[0]
                Gtk.Menu.popup(self, None, None, None, None, event.button, event.time)
            else:
                Gtk.Menu.popup(self, *args)

    def popup_at_widget(self, widget: Gtk.Widget, event: Gdk.Event) -> None:
        Gtk.Menu.popup_at_widget(
            self, widget, Gdk.Gravity.EAST, Gdk.Gravity.EAST, event
        )


class ProviderMenu(providers.ProviderHandler, Menu):
    """
    A menu that can be added to by registering a menu item with
    the providers system. If desired, a menu item can be targeted
    towards a specific parent widget.
    """

    def __init__(self, name, parent):
        providers.ProviderHandler.__init__(self, name, parent)
        Menu.__init__(self, parent)
        for p in self.get_providers():
            self.on_provider_added(p)

    def on_provider_added(self, provider):
        self.add_item(provider)

    def on_provider_removed(self, provider):
        self.remove_item(provider)

    def set_context_func(self, context_func: callable) -> None:
        self.context_func = context_func


class MultiProviderMenu(providers.MultiProviderHandler, Menu):
    """
    A menu that can be added to by registering a menu item with
    the providers system. If desired, a menu item can be targeted
    towards a specific parent widget.

    Supports retrieving menu items from multiple providers
    """

    def __init__(self, names, parent):
        providers.MultiProviderHandler.__init__(self, names, parent)
        Menu.__init__(self, parent)
        for p in self.get_providers():
            self.on_provider_added(p)

    def on_provider_added(self, provider):
        self.add_item(provider)

    def on_provider_removed(self, provider):
        self.remove_item(provider)



# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import logging
import logging.handlers
from pprint import PrettyPrinter
import os.path
import sys

__all__ = ['start_logging', 'stop_logging']


MAX_VARS_LINES = 30
MAX_LINE_LENGTH = 100


class FilterLogger(logging.Logger):
    class Filter(logging.Filter):
        def filter(self, record):
            pass_record = True

            if FilterLogger.module is not None:
                pass_record = record.name == self.module

            if FilterLogger.level != logging.NOTSET and pass_record:
                pass_record = record.levelno == self.level

            return pass_record

    module = None
    level = logging.NOTSET

    def __init__(self, name):
        logging.Logger.__init__(self, name)

        log_filter = self.Filter(name)
        log_filter.module = FilterLogger.module
        log_filter.level = FilterLogger.level
        self.addFilter(log_filter)


class SafePrettyPrinter(PrettyPrinter):
    def _repr(self, *args, **kwargs):
        try:
            return PrettyPrinter._repr(self, *args, **kwargs)
        except Exception as e:
            return "!! Cannot format: %s" % e


class VerboseExceptionFormatter(logging.Formatter):
    """
    Taken from https://word.bitly.com/post/69080588278/logging-locals
    """

    def __init__(self, log_locals_on_exception=True, *args, **kwargs):
        self._log_locals = log_locals_on_exception
        self._printer = SafePrettyPrinter(indent=2)
        super(VerboseExceptionFormatter, self).__init__(*args, **kwargs)

    def formatException(self, exc_info):
        # First get the original formatted exception.
        exc_text = super(VerboseExceptionFormatter, self).formatException(exc_info)
        if not self._log_locals or exc_info[2] is None:
            return exc_text
        # Now we're going to format and add the locals information.
        output_lines = [exc_text, '']
        tb = exc_info[2]  # This is the outermost frame of the traceback.
        while tb.tb_next:
            tb = tb.tb_next  # Zoom to the innermost frame.
        output_lines.append('Locals at innermost frame:\n')
        locals_text = self._printer.pformat(tb.tb_frame.f_locals)

        locals_lines = locals_text.split('\n')
        if len(locals_lines) > MAX_VARS_LINES:
            locals_lines = locals_lines[:MAX_VARS_LINES]
            locals_lines[-1] = '...'
        output_lines.extend(
            line[: MAX_LINE_LENGTH - 3] + '...' if len(line) > MAX_LINE_LENGTH else line
            for line in locals_lines
        )
        output_lines.append('')
        return '\n'.join(output_lines)


def start_logging(debug, quiet, debugthreads, module_filter, level_filter):
    """
    Starts logging, only should be called from xl.main
    """

    console_format = "%(levelname)-8s: %(message)s"
    console_loglevel = logging.INFO

    file_format = '%(asctime)s %(levelname)-8s: %(message)s (%(name)s)'
    file_loglevel = logging.INFO

    datefmt = "%H:%M:%S"

    if debugthreads:
        console_format = "%(threadName)s:" + console_format
    else:
        logging.logThreads = 0

    logging.logMultiprocessing = 0
    logging.logProcesses = 0

    if debug:
        file_loglevel = logging.DEBUG
        console_loglevel = logging.DEBUG
        console_format = "%(asctime)s,%(msecs)03d:" + console_format
        console_format += " (%(name)s)"  # add module name
    elif quiet:
        console_loglevel = logging.WARNING

    if debug:
        fmt_class = VerboseExceptionFormatter
    else:
        fmt_class = logging.Formatter

    # Logging to terminal
    handler = logging.StreamHandler()
    handler.setFormatter(fmt_class(fmt=console_format, datefmt=datefmt))
    logging.root.addHandler(handler)
    logging.root.setLevel(console_loglevel)

    if module_filter or level_filter:
        FilterLogger.module = module_filter
        if level_filter is not None:
            FilterLogger.level = getattr(logging, level_filter)
        logging.setLoggerClass(FilterLogger)

    # Create log directory
    from . import xdg

    logdir = xdg.get_logs_dir()

    # Logging to file; this also automatically rotates the logs
    logfile = logging.handlers.RotatingFileHandler(
        os.path.join(logdir, 'exaile.log'), mode='a', backupCount=5, delay=True
    )
    logfile.doRollover()  # each session gets its own file
    logfile.setLevel(file_loglevel)
    logfile.setFormatter(fmt_class(fmt=file_format, datefmt=datefmt))
    logging.root.addHandler(logfile)

    # GTK3 supports sys.excepthook
    if debug:
        logger = logging.getLogger('gtk')

        def log_unhandled_exception(*args):
            logger.error("Unhandled exception", exc_info=args)

        sys.excepthook = log_unhandled_exception

    # Strictly speaking, this isn't logging, but it's useful for debugging
    # when Exaile hangs.
    if sys.stderr.isatty():
        import faulthandler

        # Windows doesn't allow custom fault handler registration
        if hasattr(faulthandler, 'register'):
            import signal

            faulthandler.register(signal.SIGUSR2)
        faulthandler.enable()


def stop_logging():
    logging.shutdown()



#
# DANGER DANGER
#
# You must call Gst.init() before trying to use any GStreamer elements for
# anything. For example, if you import a module that has an object that
# inherits from Gst.Bin... that can potentially mess up any GStreamer object
# creation from that point forward.
#
# This means that Gst.init needs to be called before any plugins are loaded
#

import gi

gi.require_version('Gst', '1.0')
from gi.repository import Gst

Gst.init(None)

__gst_version__ = '%s.%s.%s' % (
    Gst.VERSION_MAJOR,
    Gst.VERSION_MINOR,
    Gst.VERSION_MICRO,
)

from xl.version import register

register('GStreamer', __gst_version__)

del register



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import locale

from gi.repository import GLib
from gi.repository import Gtk

from xl import event, main, plugins, xdg
from xlgui.widgets import common, dialogs
from xl.nls import gettext as _, ngettext

import logging

logger = logging.getLogger(__name__)

name = _('Plugins')
ui = xdg.get_data_path('ui', 'preferences', 'plugin.ui')


class PluginManager:
    """
    Gui to manage plugins
    """

    def __init__(self, preferences, builder):
        """
        Initializes the manager
        """
        self.preferences = preferences
        builder.connect_signals(self)
        self.plugins = main.exaile().plugins

        self.window = builder.get_object('PreferencesDialog')

        self.message = dialogs.MessageBar(
            parent=builder.get_object('preferences_pane'), buttons=Gtk.ButtonsType.CLOSE
        )
        self.message.connect('response', self.on_messagebar_response)

        self.list = builder.get_object('plugin_tree')
        self.enabled_cellrenderer = builder.get_object('enabled_cellrenderer')

        if main.exaile().options.Debug:
            reload_cellrenderer = common.ClickableCellRendererPixbuf()
            reload_cellrenderer.props.icon_name = 'view-refresh'
            reload_cellrenderer.props.xalign = 1
            reload_cellrenderer.connect('clicked', self.on_reload_cellrenderer_clicked)

            name_column = builder.get_object('name_column')
            name_column.pack_start(reload_cellrenderer, True)
            name_column.add_attribute(reload_cellrenderer, 'visible', 3)

        """Show trash can"""
        remove_cellrenderer = common.ClickableCellRendererPixbuf()
        remove_cellrenderer.props.icon_name = 'edit-delete'
        remove_cellrenderer.props.xalign = 1
        remove_cellrenderer.connect('clicked', self.on_remove_cellrenderer_clicked)

        name_column = builder.get_object('name_column')
        name_column.pack_start(remove_cellrenderer, True)
        name_column.add_attribute(remove_cellrenderer, 'visible', 8)

        self.version_label = builder.get_object('version_label')
        self.author_label = builder.get_object('author_label')
        self.name_label = builder.get_object('name_label')
        self.description = builder.get_object('description_view')

        self.model = builder.get_object('model')
        self.filter_model = self.model.filter_new()

        self.show_incompatible_cb = builder.get_object('show_incompatible_cb')

        self.filter_model.set_visible_func(self._model_visible_func)

        selection = self.list.get_selection()
        selection.connect('changed', self.on_selection_changed)
        self._load_plugin_list()

        self._evt_rm1 = event.add_ui_callback(
            self.on_plugin_event, 'plugin_enabled', None, True
        )
        self._evt_rm2 = event.add_ui_callback(
            self.on_plugin_event, 'plugin_disabled', None, False
        )
        self.list.connect('destroy', self.on_destroy)

        GLib.idle_add(selection.select_path, (0,))
        GLib.idle_add(self.list.grab_focus)

    def _load_plugin_list(self, select_plugin=None):
        """
        Loads the plugin list
        """
        plugins = self.plugins.list_installed_plugins()
        uncategorized = _('Uncategorized')
        plugins_dict = {uncategorized: []}
        failed_list = []

        self.plugin_to_path = {}

        for plugin_name in plugins:
            try:
                info = self.plugins.get_plugin_info(plugin_name)

                compatible = self.plugins.is_compatible(info)
                broken = self.plugins.is_potentially_broken(info)
                user_installed = self.plugins.is_user_installed(plugin_name)

            except Exception:
                logger.exception("Error loading plugin info for %r", plugin_name)
                failed_list += [plugin_name]
                continue

            # determine icon to show
            if not compatible:
                icon = 'dialog-error'
            elif broken:
                icon = 'dialog-warning'
            else:
                icon = None

            enabled = plugin_name in self.plugins.enabled_plugins
            plugin_data = (
                plugin_name,
                info['Name'],
                str(info['Version']),
                enabled,
                icon,
                broken,
                compatible,
                True,
                user_installed,
            )

            if 'Category' in info:
                cat = plugins_dict.setdefault(info['Category'], [])
                cat.append(plugin_data)
            else:
                plugins_dict[uncategorized].append(plugin_data)

        self.list.set_model(None)
        self.model.clear()

        def categorykey(item):
            if item[0] == uncategorized:
                return '\xff' * 10
            return locale.strxfrm(item[0])

        plugins_dict = sorted(plugins_dict.items(), key=categorykey)

        for category, plugins_list in plugins_dict:
            plugins_list.sort(key=lambda x: locale.strxfrm(x[1]))

            it = self.model.append(
                None, (None, category, '', False, '', False, True, False, False)
            )

            for plugin_data in plugins_list:
                pit = self.model.append(it, plugin_data)
                path = self.model.get_string_from_iter(pit)
                self.plugin_to_path[plugin_data[0]] = path

        self.list.set_model(self.filter_model)

        # TODO: Keep track of which categories are already expanded, and only expand those
        self.list.expand_all()

        if select_plugin:
            path = self.plugin_to_path[select_plugin]
            selection = self.list.get_selection()
            selection.select_path(path)
            self.list.scroll_to_cell(path)

        if failed_list:
            self.message.show_error(
                _('Could not load plugin info!'),
                ngettext('Failed plugin: %s', 'Failed plugins: %s', len(failed_list))
                % ', '.join(failed_list),
            )

    def on_destroy(self, widget):
        self._evt_rm1()
        self._evt_rm2()

    def on_messagebar_response(self, widget, response):
        """
        Hides the messagebar if requested
        """
        if response == Gtk.ResponseType.CLOSE:
            widget.hide()

    def on_plugin_tree_row_activated(self, tree, path, column):
        """
        Enables or disables the selected plugin
        """
        self.enabled_cellrenderer.emit('toggled', path[0])

    def on_remove_cellrenderer_clicked(self, cellrenderer, path: str) -> None:
        """
        Remove a user installed plugin
        """
        plugin_name = self.filter_model[path][0]
        user_installed = self.filter_model[path][8]

        if not user_installed:
            return

        GLib.idle_add(self.ask_for_remove, plugin_name)

    def ask_for_remove(self, plugin_name: str):
        response = dialogs.yesno(
            self.window,
            "\n".join(
                [
                    _('Do you really want to remove this plugin?'),
                ]
            ),
        )
        if response != Gtk.ResponseType.YES:
            return

        try:
            self.plugins.uninstall_plugin(plugin_name)
        except Exception as e:
            self.message.show_error(_('Could not remove plugin!'), str(e))
            return

        self._load_plugin_list()

    def on_reload_cellrenderer_clicked(self, cellrenderer, path):
        """
        Reloads a plugin from scratch
        """
        plugin_name = self.filter_model[path][0]
        enabled = self.filter_model[path][3]

        if enabled:
            try:
                self.plugins.disable_plugin(plugin_name)
            except Exception as e:
                self.message.show_error(_('Could not disable plugin!'), str(e))
                return

        logger.info('Reloading plugin %s...', plugin_name)
        self.plugins.load_plugin(plugin_name, reload_plugin=True)

        if enabled:
            try:
                self.plugins.enable_plugin(plugin_name)
            except Exception as e:
                self.message.show_error(_('Could not enable plugin!'), str(e))
                return

    def on_install_plugin_button_clicked(self, button):
        """
        Shows a dialog allowing the user to choose a plugin to install
        from the filesystem
        """
        dialog = Gtk.FileChooserDialog(
            title=_('Choose a Plugin'),
            transient_for=self.preferences.parent,
        )
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_ADD,
            Gtk.ResponseType.OK,
        )

        filter = Gtk.FileFilter()
        filter.set_name(_('Plugin Archives'))
        filter.add_pattern("*.exz")
        filter.add_pattern("*.tar.gz")
        filter.add_pattern("*.tar.bz2")
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name(_('All Files'))
        filter.add_pattern('*')
        dialog.add_filter(filter)

        result = dialog.run()
        dialog.hide()

        if result == Gtk.ResponseType.OK:

            def install(filename: str, overwrite_existing: bool = False) -> str:
                try:
                    plugin_name = self.plugins.install_plugin(
                        filename, overwrite_existing
                    )
                except plugins.PluginExistsError as e:
                    response = dialogs.yesno(
                        self.window,
                        "\n".join(
                            [
                                _('Plugin already exists!'),
                                _('Do you want to override it?'),
                            ]
                        ),
                    )
                    if response == Gtk.ResponseType.YES:
                        return install(filename, True)
                except plugins.InvalidPluginError as e:
                    self.message.show_error(
                        _('Plugin file installation failed!'), str(e)
                    )
                    return ''
                return plugin_name

            plugin_name = install(dialog.get_filename())
            if not plugin_name:
                return

            self._load_plugin_list(plugin_name)

    def on_selection_changed(self, selection, user_data=None):
        """
        Called when a row is selected
        """
        model, paths = selection.get_selected_rows()
        if not paths:
            return

        row = model[paths[0]]

        if not row[7]:
            self.author_label.set_label('')
            self.description.get_buffer().set_text('')
            self.name_label.set_label('')
            return

        info = self.plugins.get_plugin_info(row[0])

        self.author_label.set_label(",\n".join(info['Authors']))

        self.description.get_buffer().set_text(info['Description'].replace(r'\n', "\n"))

        self.name_label.set_markup(
            "<b>%s</b> <small>%s</small>" % (info['Name'], info['Version'])
        )

    def on_enabled_cellrenderer_toggled(self, cellrenderer, path):
        """
        Called when the checkbox is toggled
        """
        path = Gtk.TreePath.new_from_string(path)
        plugin_name = self.filter_model[path][0]
        if plugin_name is None:
            return

        enable = not self.filter_model[path][3]

        if enable:
            try:
                self.plugins.enable_plugin(plugin_name, True)
            except Exception as e:
                self.message.show_error(_('Could not enable plugin!'), str(e))
                return
        else:
            try:
                self.plugins.disable_plugin(plugin_name)
            except Exception as e:
                self.message.show_error(_('Could not disable plugin!'), str(e))
                return

        self.on_selection_changed(self.list.get_selection())

    def on_plugin_event(self, evtname, obj, plugin_name, enabled):
        if hasattr(self.plugins.loaded_plugins[plugin_name], 'get_preferences_pane'):
            self.preferences._load_plugin_pages()

        path = self.plugin_to_path[plugin_name]
        self.model[path][3] = enabled

    def on_show_incompatible_cb_toggled(self, widget):
        self.filter_model.refilter()

    def _model_visible_func(self, model, iter, data):
        row = model[iter]
        compatible = row[6]
        return compatible or self.show_incompatible_cb.get_active()


def init(preferences, xml):
    PluginManager(preferences, xml)



# Copyright (C) 2011 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from xl import event


class PlaybackAdapter:
    """
    Basic class which listens for playback changes
    """

    def __init__(self, player):
        self.__player = player
        self.__events = (
            'playback_track_start',
            'playback_track_end',
            'playback_player_end',
            'playback_toggle_pause',
            'playback_error',
        )

        for e in self.__events:
            event.add_callback(getattr(self, 'on_%s' % e), e, player)

        if player.current is not None:
            self.on_playback_track_start('playback_track_start', player, player.current)

            if player.is_paused():
                self.on_playback_toggle_pause(
                    'playback_toggle_pause', player, player.current
                )

    def destroy(self):
        """
        Cleanups
        """
        for e in self.__events:
            event.remove_callback(getattr(self, 'on_%s' % e), e, self.__player)

    def on_playback_track_start(self, event, player, track):
        """Override"""
        pass

    def on_playback_track_end(self, event, player, track):
        """Override"""
        pass

    def on_playback_player_end(self, event, player, track):
        """Override"""
        pass

    def on_playback_toggle_pause(self, event, player, track):
        """Override"""
        pass

    def on_playback_error(self, event, player, message):
        """Override"""
        pass


class QueueAdapter:
    """
    Basic class which listens for queue changes
    """

    def __init__(self, queue):
        self.__queue = queue

        event.add_callback(
            self.on_queue_current_playlist_changed,
            'queue_current_playlist_changed',
            queue,
        )
        event.add_callback(
            self.__on_playlist_current_position_changed,
            'playlist_current_position_changed',
        )
        event.add_callback(self.__on_playlist_tracks_added, 'playlist_tracks_added')
        event.add_callback(self.__on_playlist_tracks_removed, 'playlist_tracks_removed')

    def destroy(self):
        """
        Cleanups
        """
        event.remove_callback(
            self.on_queue_current_playlist_changed,
            'queue_current_playlist_changed',
            self.__queue,
        )
        event.remove_callback(
            self.__on_playlist_current_position_changed,
            'playlist_current_position_changed',
        )
        event.remove_callback(self.__on_playlist_tracks_added, 'playlist_tracks_added')
        event.remove_callback(
            self.__on_playlist_tracks_removed, 'playlist_tracks_removed'
        )

    def __on_playlist_current_position_changed(self, event, playlist, positions):
        """
        Forwards the event if emitted by the queue
        """
        if playlist is self.__queue.current_playlist:
            self.on_queue_current_position_changed(event, playlist, positions)

    def __on_playlist_tracks_added(self, event, playlist, tracks):
        """
        Forwards the event if emitted by the queue
        """
        if playlist is self.__queue.current_playlist:
            self.on_queue_tracks_added(event, playlist, tracks)

    def __on_playlist_tracks_removed(self, event, playlist, tracks):
        """
        Forwards the event if emitted by the queue
        """
        if playlist is self.__queue.current_playlist:
            self.on_queue_tracks_removed(event, playlist, tracks)

    def on_queue_current_playlist_changed(self, event, queue, playlist):
        """Override"""
        pass

    def on_queue_current_position_changed(self, event, playlist, positions):
        """Override"""
        pass

    def on_queue_tracks_added(self, event, queue, tracks):
        """Override"""
        pass

    def on_queue_tracks_removed(self, event, queue, tracks):
        """Override"""
        pass



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.
"""
This module serves as frontend for GStreamer Plugin Base Utils (GstPbutils) for
installing missing plugins.

See also:
* https://github.com/quodlibet/quodlibet/blob/master/quodlibet/quodlibet/player/gstbe/player.py#L548
* https://git.gnome.org/browse/totem/tree/src/backend/bacon-video-widget-gst-missing-plugins.c

This module must be a singleton, i.e. no instances are allowed. Thus, there is
no class.

Created on 25.04.2017

@author: Christian Stadelmann
"""

import logging

import gi

gi.require_version('GstPbutils', '1.0')
from gi.repository import GstPbutils

from xl import event
from xl.nls import gettext as _

LOGGER = logging.getLogger(__name__)

MISSING_PLUGIN_URL = "https://exaile.readthedocs.io/en/stable/user/faq.html"

# how to test:
# the helper binary is located at /usr/libexec/gstreamer/gst-install-plugins-helper


def handle_message(message, engine):
    """
    Handles `message` by checking whether it is a "missing plugin" message.
    If it is, takes all required steps to make sure that
        * playback is paused
        * the user is notified of the issue
        * the user gets a hint on which software to install

    @param message: a Gst.Message of type Gst.MessageType.Element
    @param engine: an instance of xl.player.gst.engine.ExaileGstEngine

    @return: True if the message was a "missing plugin" message and was
                being handled. This does not mean that the plugin installed
                successfully.
             False if the message should be handled by some other code
                because it is not related to a missing plugin.
    """
    if not GstPbutils.is_missing_plugin_message(message):
        return False

    __handle_plugin_missing_message(message, engine)
    return True


def __handle_plugin_missing_message(message, engine):
    desc = GstPbutils.missing_plugin_message_get_description(message)
    installer_details = GstPbutils.missing_plugin_message_get_installer_detail(message)
    LOGGER.warning("A plugin for %s is missing, stopping playback", desc)

    user_message = _(
        "A GStreamer 1.x plugin for %s is missing. "
        "Without this software installed, Exaile will not be able to play the current file. "
        "Please install the required software on your computer. See %s for details."
    ) % (desc, MISSING_PLUGIN_URL)
    # TODO make URL clickable by utilizing xlgui.widgets.dialogs.MessageBar

    engine.stop()
    __notify_user_on_error(user_message, engine)
    if GstPbutils.install_plugins_supported():
        if __run_installer_helper(installer_details):
            return
    LOGGER.warning("Installation of GStreamer plugins not supported on this platform.")


def __notify_user_on_error(message_text, engine):
    event.log_event('playback_error', engine.player, message_text)


def __run_installer_helper(installer_details):
    """

    @return True if the helper might have run. False if did not run for
                sure.
    """
    cntxt = __create_context()

    LOGGER.info("Prompting user to install missing codec(s): %s", installer_details)

    start_result = GstPbutils.install_plugins_async(
        [installer_details], cntxt, __installer_finished_callback
    )
    LOGGER.debug(
        "GstPbutils.install_plugins_async() return value: %s",
        GstPbutils.InstallPluginsReturn.get_name(start_result),
    )
    if start_result == GstPbutils.InstallPluginsReturn.INTERNAL_FAILURE:
        # should only happen when there is a bug in Exaile or its libs:
        LOGGER.error("Internal failure starting assisted GStreamer plugin installation")
        return False
    elif start_result == GstPbutils.InstallPluginsReturn.HELPER_MISSING:
        # we expect that to happen on some platforms
        LOGGER.warning("Helper missing for assisted installation of Gstreamer plugins")
        return False
    elif start_result == GstPbutils.InstallPluginsReturn.INSTALL_IN_PROGRESS:
        LOGGER.warning("Another assisted plugin installation is already in progress")
        return False
    elif start_result == GstPbutils.InstallPluginsReturn.STARTED_OK:
        LOGGER.info("Successfully started assisted GStreamer plugin installation")
        return True
    else:
        LOGGER.error(
            "Code should not be reached. "
            "Unexpected return value from install_plugins_async: %s",
            GstPbutils.InstallPluginsReturn.get_name(start_result),
        )
        return False


def __installer_finished_callback(result):
    # due to a bug in PackageKit, this function will be called immediately
    # after starting the helper: https://bugs.freedesktop.org/show_bug.cgi?id=100791
    LOGGER.debug(
        "GstPbutils.install_plugins_async() helper process exit code: %s",
        GstPbutils.InstallPluginsReturn.get_name(result),
    )

    if (
        result == GstPbutils.InstallPluginsReturn.SUCCESS
        or result == GstPbutils.InstallPluginsReturn.PARTIAL_SUCCESS
    ):
        # TODO notify user that installation of plugins was successful and ask
        # the user whether we may restart GStreamer engine to apply plugin updates,
        # because the user might have resumed playback in the meantime, and we do
        # not want to interrupt.
        #
        # engine.stop()
        # engine.destroy(permanent=False)
        # Gst.update_registry()
        # engine.initialize()
        pass
    elif result == GstPbutils.InstallPluginsReturn.NOT_FOUND:
        LOGGER.warning("GStreamer helper was unable to install missing plugin.")
        # we do not care about these, the user already got a notification how
        # to install plugins. There is nothing more we can do.
    elif (
        result == GstPbutils.InstallPluginsReturn.ERROR
        or result == GstPbutils.InstallPluginsReturn.CRASHED
        or result == GstPbutils.InstallPluginsReturn.INVALID
        or result == GstPbutils.InstallPluginsReturn.INTERNAL_FAILURE
    ):
        LOGGER.error(
            "GStreamer plugin helper failed with %s",
            GstPbutils.InstallPluginsReturn.get_name(result),
        )
    elif result == GstPbutils.InstallPluginsReturn.USER_ABORT:
        LOGGER.info("User aborted the GStreamer plugin installation.")
        # the user decided not to install any software, which might have been on purpose,
        # so we do not want to ask again immediately.
    else:
        LOGGER.error(
            "Code should not be reached. "
            "Unexpected return value from install_plugins_async callback: %s",
            GstPbutils.InstallPluginsReturn.get_name(result),
        )


def __create_context():
    LOGGER.info("Initializing connector for GstPbutils...")
    cntxt = GstPbutils.InstallPluginsContext()
    cntxt.set_confirm_search(True)

    # See https://standards.freedesktop.org/desktop-entry-spec/latest/ape.html
    cntxt.set_desktop_id("exaile.desktop")
    # TODO
    # cntxt.set_startup_notification_id()
    # cntxt.set_xid()
    return cntxt



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Code based on Ingelrest FranÃ§ois' "im status" plugin

import dbus
from xl import event, player

##############################################################################


class Pidgin:
    def __init__(self, dbusInterface):
        """
        Constructor
        """
        self.purple = dbusInterface

    def listAccounts(self):
        """
        Purple merges all accounts, so we return a default one
        Each account is associated with:
            * A boolean -> True if the status of this account was changed on the previous track change
        """
        return {'GenericAccount': False}

    def setStatus(self, status, attr, value):
        # this doesn't actually work, for some reason the getter always return ""
        if self.purple.PurpleStatusGetAttrString(status, attr) != value:
            self.purple.PurpleStatusSetAttrString(status, attr, value)
            return True
        return False

    def setTune(self, artist, title, album):
        """
        Change the tune status
        Return True if the message is successfully updated
        """
        accounts = self.purple.PurpleAccountsGetAll()

        for account in accounts:
            if not self.purple.PurpleAccountIsConnected(account):
                continue
            p = self.purple.PurpleAccountGetPresence(account)
            status = self.purple.PurplePresenceGetStatus(p, "tune")

            if status != 0:
                updated = False
                if len(title) + len(artist) + len(album) == 0:
                    if self.purple.PurpleStatusIsActive(status):
                        self.purple.PurpleStatusSetActive(status, False)
                else:
                    self.purple.PurpleStatusSetActive(status, True)
                    updated |= self.setStatus(status, "tune_title", title)
                    updated |= self.setStatus(status, "tune_artist", artist)
                    updated |= self.setStatus(status, "tune_album", album)
                if updated:
                    active = self.purple.PurplePresenceGetActiveStatus(p)
                    self.purple.PurplePrplChangeAccountStatus(account, active, status)

        return True


##############################################################################


def on_begin_action(type, player, track):
    client.setTune(
        track.get_tag_display('artist'),
        track.get_tag_display('title'),
        track.get_tag_display('album'),
    )


def on_stop_action(type, player, track):
    client.setTune("", "", "")


def on_pause_action(type, player, track):
    if player.is_playing():
        on_begin_action(type, player, track)
    else:
        on_stop_action(type, player, track)


def enable(exaile):
    global client
    obj = dbus.SessionBus().get_object(
        'im.pidgin.purple.PurpleService', '/im/pidgin/purple/PurpleObject'
    )
    purple = dbus.Interface(obj, "im.pidgin.purple.PurpleInterface")
    client = Pidgin(purple)
    event.add_callback(on_stop_action, 'quit_application')
    event.add_callback(on_stop_action, 'playback_player_end', player.PLAYER)
    event.add_callback(on_begin_action, 'playback_track_start', player.PLAYER)
    event.add_callback(on_pause_action, 'playback_toggle_pause', player.PLAYER)


def disable(exaile):
    event.remove_callback(on_stop_action, 'quit_application')
    event.remove_callback(on_stop_action, 'playback_player_end', player.PLAYER)
    event.remove_callback(on_begin_action, 'playback_track_start', player.PLAYER)
    event.remove_callback(on_pause_action, 'playback_toggle_pause', player.PLAYER)



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from datetime import datetime, timedelta
import os
import re
import zlib
import threading

from xl.nls import gettext as _
from xl.trax import Track
from xl import common, event, providers, settings, xdg


class LyricsNotFoundException(Exception):
    pass


class LyricsCache:
    """
    Basically just a thread-safe shelf for convenience.
    Supports container syntax.
    """

    def __init__(self, location, default=None):
        """
        @param location: specify the shelve file location

        @param default: can specify a default to return from getter when
            there is nothing in the shelve
        """
        self.location = location
        self.db = common.open_shelf(location)
        self.lock = threading.Lock()
        self.default = default

        # Callback to close db
        event.add_callback(self.on_quit_application, 'quit_application')

    def on_quit_application(self, *args):
        """
        Closes db on quit application
        Gets the lock/wait operations
        """
        with self.lock:
            self.db.close()

    def keys(self):
        """
        Return the shelve keys
        """
        return self.db.keys()

    def _get(self, key, default=None):
        with self.lock:
            try:
                return self.db[key]
            except Exception:
                return default if default is not None else self.default

    def _set(self, key, value):
        with self.lock:
            self.db[key] = value
            # force save, wasn't auto-saving...
            self.db.sync()

    def __getitem__(self, key):
        return self._get(key)

    def __setitem__(self, key, value):
        self._set(key, value)

    def __contains__(self, key):
        return key in self.db

    def __delitem__(self, key):
        with self.lock:
            del self.db[key]

    def __iter__(self):
        return self.db.__iter__()

    def __len__(self):
        return len(self.db)


class LyricsManager(providers.ProviderHandler):
    """
    Lyrics Manager

    Manages talking to the lyrics plugins and updating the track
    """

    def __init__(self):
        providers.ProviderHandler.__init__(self, "lyrics")
        self.preferred_order = settings.get_option('lyrics/preferred_order', [])
        self.cache = LyricsCache(os.path.join(xdg.get_cache_dir(), 'lyrics.cache'))

        event.add_callback(self.on_track_tags_changed, 'track_tags_changed')

    def __get_cache_key(self, track: Track, provider) -> str:
        """
        Returns the cache key for a specific track and lyrics provider

        :param track: a track
        :param provider: a lyrics provider
        :return: the appropriate cache key
        """
        return (
            track.get_loc_for_io()
            + provider.display_name
            + track.get_tag_display('artist')
            + track.get_tag_display('title')
        )

    def set_preferred_order(self, order):
        """
        Sets the preferred search order

        :param order: a list containing the order you'd like to search
            first
        """
        if not type(order) in (list, tuple):
            raise AttributeError("order must be a list or tuple")
        self.preferred_order = order
        settings.set_option('lyrics/preferred_order', list(order))

    def find_lyrics(self, track, refresh=False):
        """
        Fetches lyrics for a track either from
            1. a backend lyric plugin
            2. the actual tags in the track

        :param track: the track we want lyrics for, it
            must have artist/title tags

        :param refresh: if True, try to refresh cached data even if
            not expired

        :return: tuple of the following format (lyrics, source, url)
            where lyrics are the lyrics to the track
            source is where it came from (file, lyrics wiki,
            lyrics fly, etc.)
            url is a link to the lyrics (where applicable)

        :raise LyricsNotFoundException: when lyrics are not
            found
        """
        lyrics = None
        source = None
        url = None

        for method in self.get_providers():
            try:
                lyrics, source, url = self._find_cached_lyrics(method, track, refresh)
            except LyricsNotFoundException:
                continue
            break
        else:
            # This only happens if all providers raised LyricsNotFoundException.
            raise LyricsNotFoundException()

        lyrics = lyrics.strip()

        return (lyrics, source, url)

    def find_all_lyrics(self, track, refresh=False):
        """
        Like find_lyrics but fetches all sources and returns
        a list of lyrics.

        :param track: the track we want lyrics for, it
            must have artist/title tags

        :param refresh: if True, try to refresh cached data even if
            not expired

        :return: list of tuples in the same format as
            find_lyrics's return value

        :raise LyricsNotFoundException: when lyrics are not
            found from all sources.
        """
        lyrics_found = []

        for method in self.get_providers():
            lyrics = None
            source = None
            url = None
            try:
                lyrics, source, url = self._find_cached_lyrics(method, track, refresh)
            except LyricsNotFoundException:
                continue
            lyrics = lyrics.strip()
            lyrics_found.append((method.display_name, lyrics, source, url))

        if not lyrics_found:
            # no lyrics were found, raise an exception
            raise LyricsNotFoundException()

        return lyrics_found

    def _find_cached_lyrics(self, method, track, refresh=False):
        """
        Checks the cache for lyrics.  If found and not expired, returns
        cached results, otherwise tries to fetch from method.

        :param method: the LyricSearchMethod to fetch lyrics from.

        :param track: the track we want lyrics for, it
            must have artist/title tags

        :param refresh: if True, try to refresh cached data even if
            not expired

        :return: list of tuples in the same format as
            find_lyric's return value

        :raise LyricsNotFoundException: when lyrics are not found
            in cache or fetched from method
        """
        lyrics = None
        source = None
        url = None
        cache_time = settings.get_option('lyrics/cache_time', 720)  # in hours
        key = self.__get_cache_key(track, method)

        # check cache for lyrics
        if key in self.cache:
            lyrics, source, url, time = self.cache[key]
            # return if they are not expired
            now = datetime.now()
            if now - time < timedelta(hours=cache_time) and not refresh:
                try:
                    lyrics = zlib.decompress(lyrics)
                except zlib.error as e:
                    raise LyricsNotFoundException(e)
                return (lyrics.decode('utf-8', errors='replace'), source, url)

        lyrics, source, url = method.find_lyrics(track)
        assert isinstance(lyrics, str), (method, track)

        # update cache
        time = datetime.now()
        self.cache[key] = (zlib.compress(lyrics.encode('utf-8')), source, url, time)

        return (lyrics, source, url)

    def on_provider_removed(self, provider):
        """
        Remove the provider from the methods dict, and the
        preferred_order dict if needed.

        :param provider: the provider instance being removed.
        """
        try:
            self.preferred_order.remove(provider.name)
        except (ValueError, AttributeError):
            pass

    def on_track_tags_changed(self, e, track, tags):
        """
        Updates the internal cache upon lyric tag changes
        """
        if 'lyrics' in tags:
            local_provider = self.get_provider('__local')

            # If the local tag provider was removed, don't bother
            if local_provider is None:
                return

            key = self.__get_cache_key(track, local_provider)

            # Try to remove the corresponding cache entry
            try:
                del self.cache[key]
            except KeyError:
                pass


MANAGER = LyricsManager()


class LyricSearchMethod:
    """
    Lyrics plugins will subclass this
    """

    def find_lyrics(self, track):
        """
        Called by LyricsManager when lyrics are requested

        :param track: the track that we want lyrics for
        :return: tuple of lyrics text, provider name, URL
        :rtype: Tuple[unicode, basestring, basestring]
        :raise: LyricsNotFoundException if not found
        """
        raise NotImplementedError

    def _set_manager(self, manager):
        """
        Sets the lyrics manager.

        Called when this method is added to the lyrics manager.

        :param manager: the lyrics manager
        """
        self.manager = manager

    def remove_script(self, data):
        p = re.compile(r'<script.*/script>')
        return p.sub('', data)

    def remove_div(self, data):
        p = re.compile(r'<div.*/div>')
        return p.sub('', data)

    def remove_html_tags(self, data):
        data = data.replace('<br/>', '\n')
        p = re.compile(r'<[^<]*?/?>')
        data = p.sub('', data)
        p = re.compile(r'/<!--.*?-->/')
        return p.sub('', data)


class LocalLyricSearch(LyricSearchMethod):
    name = "__local"
    display_name = _("Local")

    def find_lyrics(self, track):
        lyrics = track.get_tag_disk('lyrics')
        if not lyrics:
            raise LyricsNotFoundException()
        return (lyrics[0], self.name, "")


providers.register('lyrics', LocalLyricSearch())



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Arunas Radzvilavicius, arunas.rv@gmail.com

from gi.repository import Gtk
from gi.repository import Pango


class AboutWindow:
    def __init__(self):
        self.book = None
        self.showing = False

        self.win = Gtk.Window()
        self.win.set_title("About")
        self.win.set_default_size(300, 200)
        self.win.set_geometry_hints(self.win, min_width=100, min_height=100)
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.win.add(self.vbox)

        self.scrollwin = Gtk.ScrolledWindow()
        self.scrollwin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.vbox.pack_start(self.scrollwin, True, True, 0)

        self.textview = Gtk.TextView()
        self.textview.set_cursor_visible(False)
        self.textview.set_editable(False)
        self.textview.set_wrap_mode(Gtk.WrapMode.WORD)
        self.textview.set_justification(Gtk.Justification.LEFT)
        self.textview.set_left_margin(4)
        self.textview.set_left_margin(4)

        self.scrollwin.add(self.textview)

        self.textbuffer = Gtk.TextBuffer()
        self.textbuffer.create_tag('bold', weight=Pango.Weight.BOLD)
        self.textview.set_buffer(self.textbuffer)

        self.hbox = Gtk.Box()
        self.vbox.pack_start(self.hbox, False, False, 2)

        self.closebutton = Gtk.Button("Close")
        self.closebutton.connect("pressed", self.closebutton_pressed)
        self.closeimage = Gtk.Image.new_from_icon_name(
            'window-close', Gtk.IconSize.MENU
        )
        self.closebutton.set_image(self.closeimage)
        self.hbox.pack_end(self.closebutton, False, False)

        self.vbox.show_all()

        self.win.connect("delete-event", self.on_delete)

    def set_text(self, book):
        self.book = book
        titlelength = len(book.title)
        if book.info is None:
            book.info = "No information."
        self.textbuffer.set_text(book.title + '\n\n' + book.info + '\n')
        start = self.textbuffer.get_iter_at_offset(0)
        end = self.textbuffer.get_iter_at_offset(titlelength)
        self.textbuffer.apply_tag_by_name('bold', start, end)

    def on_delete(self, window, event):
        self.win.hide()
        self.showing = False
        return True

    def closebutton_pressed(self, widget):
        self.win.hide()
        self.showing = False



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging

from gi.repository import Gdk
from gi.repository import Gtk

from xl import event, player, providers, settings
from xl.nls import gettext as _
from xlgui.widgets.info import TrackToolTip
from xlgui.widgets import menu, menuitems, playlist, playback
from xlgui import guiutil

logger = logging.getLogger(__name__)


def is_supported():
    """
    On some platforms (e.g. Linux+Wayland) tray icons are not supported.
    """
    supported = not guiutil.platform_is_wayland()

    if not supported:
        logger.debug("No tray icon support on this platform")

    return supported


def __create_tray_context_menu():
    sep = menu.simple_separator
    items = []
    # Play/Pause
    items.append(
        playback.PlayPauseMenuItem('playback-playpause', player.PLAYER, after=[])
    )
    # Next
    items.append(
        playback.NextMenuItem('playback-next', player.PLAYER, after=[items[-1].name])
    )
    # Prev
    items.append(
        playback.PrevMenuItem('playback-prev', player.PLAYER, after=[items[-1].name])
    )
    # Stop
    items.append(
        playback.StopMenuItem('playback-stop', player.PLAYER, after=[items[-1].name])
    )
    # ----
    items.append(sep('playback-sep', [items[-1].name]))
    # Shuffle
    items.append(
        playlist.ShuffleModesMenuItem('playlist-mode-shuffle', after=[items[-1].name])
    )
    # Repeat
    items.append(
        playlist.RepeatModesMenuItem('playlist-mode-repeat', after=[items[-1].name])
    )
    # Dynamic
    items.append(
        playlist.DynamicModesMenuItem('playlist-mode-dynamic', after=[items[-1].name])
    )
    # ----
    items.append(sep('playlist-mode-sep', [items[-1].name]))
    # Rating

    def rating_get_tracks_func(parent, context):
        current = player.PLAYER.current
        if current:
            return [current]
        else:
            return []

    items.append(
        menuitems.RatingMenuItem('rating', [items[-1].name], rating_get_tracks_func)
    )
    # Remove
    items.append(playlist.RemoveCurrentMenuItem([items[-1].name]))
    # ----
    items.append(sep('misc-actions-sep', [items[-1].name]))
    # Quit

    def quit_cb(*args):
        from xl import main

        main.exaile().quit()

    items.append(
        menu.simple_menu_item(
            'quit-application',
            [items[-1].name],
            _("_Quit Exaile"),
            'application-exit',
            callback=quit_cb,
        )
    )
    for item in items:
        providers.register('tray-icon-context', item)


__create_tray_context_menu()


class BaseTrayIcon:
    """
    Trayicon base, needs to be derived from
    """

    def __init__(self, main):
        self.main = main
        self.VOLUME_STEP = 0.05

        self.tooltip = TrackToolTip(self, player.PLAYER)
        self.tooltip.set_auto_update(True)
        self.tooltip.set_display_progress(True)

        self.menu = menu.ProviderMenu('tray-icon-context', self)
        self.update_icon()
        self.connect_events()
        event.log_event('tray_icon_toggled', self, True)

    def destroy(self):
        """
        Unhides the window and removes the tray icon
        """
        # FIXME: Allow other windows too
        if not self.main.window.get_property('visible'):
            self.main.window.deiconify()
            self.main.window.present()

        self.disconnect_events()
        self.set_visible(False)
        self.tooltip.destroy()
        event.log_event('tray_icon_toggled', self, False)

    def connect_events(self):
        """
        Connects various callbacks with events
        """
        self.connect('button-press-event', self.on_button_press_event)
        self.connect('scroll-event', self.on_scroll_event)

        event.add_ui_callback(
            self.on_playback_change_state, 'playback_player_end', player.PLAYER
        )
        event.add_ui_callback(
            self.on_playback_change_state, 'playback_track_start', player.PLAYER
        )
        event.add_ui_callback(
            self.on_playback_change_state, 'playback_toggle_pause', player.PLAYER
        )
        event.add_ui_callback(
            self.on_playback_change_state, 'playback_error', player.PLAYER
        )

    def disconnect_events(self):
        """
        Disconnects various callbacks from events
        """
        event.remove_callback(
            self.on_playback_change_state, 'playback_player_end', player.PLAYER
        )
        event.remove_callback(
            self.on_playback_change_state, 'playback_track_start', player.PLAYER
        )
        event.remove_callback(
            self.on_playback_change_state, 'playback_toggle_pause', player.PLAYER
        )
        event.remove_callback(
            self.on_playback_change_state, 'playback_error', player.PLAYER
        )

    def update_icon(self):
        """
        Updates icon appearance based
        on current playback state
        """
        if player.PLAYER.current is None:
            self.set_from_icon_name('exaile')
            self.set_tooltip(_('Exaile Music Player'))
        elif player.PLAYER.is_paused():
            self.set_from_icon_name('exaile-pause')
        else:
            self.set_from_icon_name('exaile-play')

    def set_from_icon_name(self, icon_name):
        """
        Updates the tray icon
        """
        pass

    def set_tooltip(self, tooltip_text):
        """
        Updates the tray icon tooltip
        """
        pass

    def set_visible(self, visible):
        """
        Shows or hides the tray icon
        """
        pass

    def on_button_press_event(self, widget, event):
        """
        Toggles main window visibility and
        pause as well as opens the context menu
        """
        if event.button == Gdk.BUTTON_PRIMARY:
            self.main.toggle_visible(bringtofront=True)
        if event.button == Gdk.BUTTON_MIDDLE:
            playback.playpause(player.PLAYER)
        if event.triggers_context_menu():
            self.menu.popup_at_pointer(event)

    def on_scroll_event(self, widget, event):
        """
        Changes volume and skips tracks on scroll
        """
        if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
            if event.direction == Gdk.ScrollDirection.UP:
                player.QUEUE.prev()
            elif event.direction == Gdk.ScrollDirection.DOWN:
                player.QUEUE.next()
        else:
            if event.direction == Gdk.ScrollDirection.UP:
                volume = settings.get_option('player/volume', 1)
                settings.set_option('player/volume', min(volume + self.VOLUME_STEP, 1))
                return True
            elif event.direction == Gdk.ScrollDirection.DOWN:
                volume = settings.get_option('player/volume', 1)
                settings.set_option('player/volume', max(0, volume - self.VOLUME_STEP))
                return True
            elif event.direction == Gdk.ScrollDirection.LEFT:
                player.QUEUE.prev()
            elif event.direction == Gdk.ScrollDirection.RIGHT:
                player.QUEUE.next()

    def on_playback_change_state(self, event, player, current):
        """
        Updates tray icon appearance
        on playback state change
        """
        self.update_icon()


class TrayIcon(Gtk.StatusIcon, BaseTrayIcon):
    """
    Wrapper around GtkStatusIcon
    """

    def __init__(self, main):
        Gtk.StatusIcon.__init__(self)
        BaseTrayIcon.__init__(self, main)

    def set_tooltip(self, tooltip_text):
        """
        Updates the tray icon tooltip
        """
        self.set_tooltip_text(tooltip_text)



# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2014-2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


class ExaileEngine:
    """
    Interface that the ExailePlayer uses to control the engines. Other
    parts of Exaile should never interact directly with the engines, but
    interact with the ExailePlayer instead.

    Engines should not call play/stop/pause/unpause on itself directly,
    should call through the player object instead.

    All public functions are assumed to be called from the Glib main
    thread, or bad things will happen.

    The ExailePlayer object provides a common API that engines must use
    to indicate the current playback status, or retrieve things such as
    the next track to be played.

    Player API: Playback status:

    * engine_notify_track_start: Called when a track begins playback
    * engine_notify_track_end: Called when a track has stopped playing
    * engine_notify_error: Call this to notify the user of an error

    Player API: Auto-advance

    * engine_autoadvance_get_next_track: Call this to determine whether
      a new track should be played after the current track
    * engine_autoadvance_notify_next: Call when the auto advance track is
      actually played.

    Player API: misc

    * engine_notify_user_volume_change: Call this when something inside
      of the engine causes the user volume to change.

    Engines are expected to honor the value of the __startoffset and
    __stopoffset flags to start/stop tracks at the appropriate offsets.
    """

    def __init__(self, name, player):
        self.name = name
        self.player = player

    def initialize(self):
        """
        Initialize the engine.

        The engine should call `player.engine_load_volume` when
        initialization has been completed.
        """
        raise NotImplementedError

    def destroy(self):
        """
        Closes all resources associated with this engine
        """
        raise NotImplementedError

    def get_current_track(self):
        """
        :returns: the current track that is playing, or None
        """
        raise NotImplementedError

    def get_position(self):
        """
        Gets the current playback position of the playing track

        :returns: the playback position in nanoseconds
        :rtype: int
        """
        raise NotImplementedError

    def get_state(self):
        """
        Gets the player state

        :returns: one of *playing*, *paused* or *stopped*
        :rtype: string
        """
        raise NotImplementedError

    def get_volume(self):
        """
        Gets the current engine volume

        :returns: the volume percentage (0..1)
        :type: float
        """
        raise NotImplementedError

    def on_track_stopoffset_changed(self, track):
        """
        Called when the stop offset of a track has been changed by the
        user. Engines should adjust the playing time of the track.
        """

    def pause(self):
        """
        Pauses the playback, does not toggle it
        """
        raise NotImplementedError

    def play(self, track, start_at, paused):
        """
        Starts the playback with the provided track or stops the playback
        immediately if none. If the __startoffset tag is set, then
        start_at will be set to that value.

        Calls self.player.engine_notify_track_end if a prior track was playing.

        Calls self.player.engine_notify_track_start when the track has started.

        :param track: the track to play
        :type track: :class:`xl.trax.Track`
        :param start_at: The offset to start playback at, in seconds; or None
        :param paused: If True, start the track in 'paused' mode
        """
        raise NotImplementedError

    def seek(self, value):
        """
        Seek to a position in the currently playing stream

        :param value: the position in seconds
        :type value: float
        """
        raise NotImplementedError

    def set_volume(self, volume):
        """
        Sets the volume on this engine

        :param volume: the volume percentage (0..1)
        :type volume: float
        """
        raise NotImplementedError

    def stop(self):
        """
        Stops playback, and calls self.player.engine_notify_track_end
        """
        raise NotImplementedError

    def unpause(self):
        """
        Unpauses the playback, does not toggle it
        """
        raise NotImplementedError



from unittest.mock import patch

from gi.repository import Gio

import xl.collection
import xl.trax.search
import xl.trax.track
import xl.trax.util


def test_is_valid_track_valid(test_track):
    fname = test_track.filename
    assert xl.trax.util.is_valid_track(fname), fname


def test_is_valid_track_invalid():
    assert not xl.trax.util.is_valid_track('/')
    assert not xl.trax.util.is_valid_track('/tmp')
    assert not xl.trax.util.is_valid_track(__file__)
    assert not xl.trax.util.is_valid_track('http:///tmp')


class TestGetTracksFromUri:
    def test_invalid(self):
        uri = __file__  # Not a URI, should fail
        assert xl.trax.util.get_tracks_from_uri(uri) == []
        uri = Gio.File.new_for_path('__nonexistent_file__').get_uri()
        assert xl.trax.util.get_tracks_from_uri(uri) == []

    def test_single(self):
        uri = Gio.File.new_for_path(__file__).get_uri()
        assert xl.trax.util.get_tracks_from_uri(uri) == [xl.trax.track.Track(uri)]

    def test_directory(self):
        uri = Gio.File.new_for_path(__file__).get_parent()
        assert uri
        uri = uri.get_uri()

        with patch('xl.collection.Library.rescan'):
            xl.trax.util.get_tracks_from_uri(uri)


class TestSortTracks:
    def setup_method(self):
        self.tracks = [
            xl.trax.track.Track(url) for url in ('/tmp/foo', '/tmp/bar', '/tmp/baz')
        ]
        for track, val in zip(self.tracks, 'aab'):
            track.set_tag_raw('artist', val)
        for track, val in zip(self.tracks, '212'):
            track.set_tag_raw('discnumber', val)
        self.fields = ('artist', 'discnumber')
        self.result = [self.tracks[1], self.tracks[0], self.tracks[2]]

    def test_sorted(self):
        assert xl.trax.util.sort_tracks(self.fields, self.tracks) == self.result

    def test_reversed(self):
        assert xl.trax.util.sort_tracks(self.fields, self.tracks, reverse=True) == list(
            reversed(self.result)
        )


class TestSortResultTracks:
    def setup_method(self):
        tracks = [
            xl.trax.track.Track(url) for url in ('/tmp/foo', '/tmp/bar', '/tmp/baz')
        ]
        for track, val in zip(tracks, 'aab'):
            track.set_tag_raw('artist', val)
        for track, val in zip(tracks, '212'):
            track.set_tag_raw('discnumber', val)
        self.tracks = [xl.trax.search.SearchResultTrack(track) for track in tracks]
        self.fields = ('artist', 'discnumber')
        self.result = [self.tracks[1], self.tracks[0], self.tracks[2]]

    def test_sorted(self):
        assert xl.trax.util.sort_result_tracks(self.fields, self.tracks) == self.result

    def test_reversed(self):
        assert xl.trax.util.sort_result_tracks(self.fields, self.tracks, True) == list(
            reversed(self.result)
        )



# Copyright (C) 2025  Johannes Sasongko <johannes sasongko org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import os.path
import sqlite3
import tempfile

import pytest

from xl import sqlitedbm


def test_needs_rw_to_write():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        db = sqlitedbm.SqliteDbm(dbpath, mode='rwc')
        db[b'key1'] = b'value1'
        del db
        db = sqlitedbm.SqliteDbm(dbpath, mode='rw')
        db[b'key2'] = b'value2'
        del db
        db = sqlitedbm.SqliteDbm(dbpath, mode='ro')
        with pytest.raises(sqlite3.OperationalError):
            db[b'key3'] = b'value3'


def test_needs_rwc_to_create():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        with pytest.raises(sqlite3.OperationalError):
            sqlitedbm.SqliteDbm(dbpath, mode='ro')
        with pytest.raises(sqlite3.OperationalError):
            sqlitedbm.SqliteDbm(dbpath, mode='rw')


def test_no_autocommit():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        db = sqlitedbm.SqliteDbm(dbpath, autocommit=False)
        db[b'key1'] = b'value1'
        del db
        # Should still flush on close.
        db = sqlitedbm.SqliteDbm(dbpath, mode='ro')
        assert db[b'key1'] == b'value1'


def test_operations():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        db = sqlitedbm.SqliteDbm(dbpath)
        assert len(db) == 0
        db[b'key1'] = b'value1'
        db[b'key2'] = b'value2'
        assert b'key1' in db
        assert db[b'key1'] == b'value1'
        del db[b'key1']
        assert len(db) == 1
        assert list(db.items()) == [(b'key2', b'value2')]
        db.close()
        with pytest.raises(Exception):
            len(db)


def test_persistence():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        db = sqlitedbm.SqliteDbm(dbpath)
        db[b'key1'] = b'value1'
        del db
        db = sqlitedbm.SqliteDbm(dbpath, mode='ro')
        assert list(db.items()) == [(b'key1', b'value1')]


def test_context_manager():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        with sqlitedbm.SqliteDbm(dbpath, mode='rwc') as db:
            pass
        with pytest.raises(Exception):
            len(db)



# Copyright (C) 2013-2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import logging

import ctypes.wintypes as cwin
import ctypes

from gi.repository import Gst

logger = logging.getLogger(__name__)


def get_priority_booster():
    """
    This hack allows us to boost the priority of GStreamer task threads on
    Windows. See https://github.com/exaile/exaile/issues/76 and
    https://bugzilla.gnome.org/show_bug.cgi?id=781998
    """
    avrt_dll = ctypes.windll.LoadLibrary("avrt.dll")
    AvSetMmThreadCharacteristics = avrt_dll.AvSetMmThreadCharacteristicsW
    AvSetMmThreadCharacteristics.argtypes = [cwin.LPCWSTR, ctypes.POINTER(cwin.DWORD)]
    AvSetMmThreadCharacteristics.restype = cwin.HANDLE

    AvRevertMmThreadCharacteristics = avrt_dll.AvRevertMmThreadCharacteristics
    AvRevertMmThreadCharacteristics.argtypes = [cwin.HANDLE]
    AvRevertMmThreadCharacteristics.restype = cwin.BOOL

    def on_stream_status(bus, message):
        """
        Called synchronously from GStreamer processing threads -- do what
        we need to do and then get out ASAP
        """
        status = message.parse_stream_status()

        # A gstreamer thread starts
        if status.type == Gst.StreamStatusType.ENTER:
            obj = message.get_stream_status_object()

            # note that we use "Pro Audio" because it gives a higher priority, and
            # that's what Chrome does anyways...
            unused = cwin.DWORD()
            obj.task_handle = AvSetMmThreadCharacteristics(
                "Pro Audio", ctypes.byref(unused)
            )

        # A gstreamer thread ends
        elif status.type == Gst.StreamStatusType.LEAVE:
            obj = message.get_stream_status_object()
            task_handle = getattr(obj, 'task_handle', None)
            if task_handle:
                AvRevertMmThreadCharacteristics(task_handle)

    def attach_priority_hook(player):
        bus = player.get_bus()
        bus.connect('sync-message::stream-status', on_stream_status)
        bus.enable_sync_message_emission()

    return attach_priority_hook



# Copyright (C) 2009-2010 Aren Olson, Johannes Schwarz
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import json
from urllib.parse import quote_plus
from xl import common, covers, event, providers

# Last.fm API Key for Exaile
# if you reuse this code in a different application, please
# register your own key with last.fm
API_KEY = '3599c79a97fd61ce518b75922688bc38'

API_ROOT_URL = 'https://ws.audioscrobbler.com/2.0/'
LASTFM = None


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, "exaile_loaded")
    else:
        _enable(None, exaile, None)


def _enable(eventname, exaile, nothing):
    global LASTFM
    LASTFM = LastFMCoverSearch(exaile)
    providers.register('covers', LASTFM)


def disable(exaile):
    providers.unregister('covers', LASTFM)


class LastFMCoverSearch(covers.CoverSearchMethod):
    """
    Searches Last.fm for covers
    """

    name = 'lastfm'
    title = 'Last.fm'
    type = 'remote'  # fetches remotely as opposed to locally

    url = '{api_rurl}?method={type}.search&{type}={value}&format=json&api_key={api_key}'

    def __init__(self, exaile):
        self.user_agent = exaile.get_user_agent_string('lastfmcovers')

    def find_covers(self, track, limit=-1):
        """
        Searches last.fm for album covers
        """
        # TODO: handle multi-valued fields better
        try:
            artist, album, title = (
                track.get_tag_raw('artist')[0],
                track.get_tag_raw('album')[0],
                track.get_tag_raw('title')[0],
            )
        except TypeError:
            return []

        if not artist or not album or not title:
            return []

        for type, value in (('album', album), ('track', title)):
            url = self.url.format(
                api_rurl=API_ROOT_URL,
                type=type,
                value=quote_plus(value.encode("utf-8")),
                api_key=API_KEY,
            )
            try:
                data = common.get_url_contents(url, self.user_agent)
            except IOError:
                return []

            try:
                data_json = json.loads(data)
            except ValueError:
                continue

            try:
                for element in data_json['results']["%smatches" % type][type]:
                    if element['artist'] == artist.encode("utf-8"):
                        for image in element['image']:
                            if image['size'] == 'extralarge':
                                return [image['#text']]
            except KeyError:
                continue

        return []

    def get_cover_data(self, cover_url):
        if not cover_url:
            return None
        try:
            return common.get_url_contents(cover_url, self.user_agent)
        except IOError:
            return None



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from copy import deepcopy
import logging
from time import time
from typing import Dict, Iterable, Iterator, List, Optional, Tuple

from xl import common, event
from xl.nls import gettext as _
from xl.trax.track import Track

logger = logging.getLogger(__name__)


class TrackHolder:
    def __init__(self, track, key, **kwargs):
        self._track = track
        self._key = key
        self._attrs = kwargs

    def __getattr__(self, attr):
        return getattr(self._track, attr)


class TrackDBIterator:
    def __init__(self, track_iterator: Iterator[Tuple[str, TrackHolder]]):
        self.iter = track_iterator

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iter)[1]._track


class TrackDB:
    """
    Manages a track database.

    Allows you to add, remove, retrieve, search, save and load
    Track objects.

    :param name:   The name of this :class:`TrackDB`.
    :param location:   Path to a file where this :class:`TrackDB`
            should be stored.
    :param pickle_attrs:   A list of attributes to store in the
            pickled representation of this object. All
            attributes listed must be built-in types, with
            one exception: If the object contains the phrase
            'tracks' in its name it may be a list or dict
            of :class:`Track` objects.
    :param load_first: Set to True if this collection should be
            loaded before any tracks are created.
    """

    def __init__(
        self,
        name: str = "",
        location: str = "",
        pickle_attrs: List[str] = [],
        loadfirst: bool = False,
    ):
        """
        Sets up the trackDB.
        """

        # ensure that the DB is always loaded before any tracks are,
        # otherwise internal values are not loaded and may be lost/corrupted
        if loadfirst and Track._get_track_count() != 0:
            raise RuntimeError(
                (
                    "Internal error! %d tracks already loaded, "
                    + "TrackDB must be loaded first!"
                )
                % Track._get_track_count()
            )

        self.name = name
        self.location = location
        self._dirty = False
        self.tracks: Dict[str, TrackHolder] = {}  # key is URI of the track
        self.pickle_attrs = pickle_attrs
        self.pickle_attrs += ['tracks', 'name', '_key']
        self._saving = False
        #: Number to use for the next `tracks-*` database key
        self._key = 0
        self._dbversion = 2.0
        self._deleted_keys = []
        if location:
            self.load_from_location()
            self._timeout_save()

    def __iter__(self):
        """
        Provide the ability to iterate over a TrackDB.
        Just as with a dictionary, if tracks are added
        or removed during iteration, iteration will halt
        with a RuntimeError.
        """
        track_iterator = iter(self.tracks.items())
        iterator = TrackDBIterator(track_iterator)
        return iterator

    def __len__(self):
        """
        Obtain a count of how many items are in the TrackDB
        """
        return len(self.tracks)

    @common.glib_wait_seconds(300)
    def _timeout_save(self):
        """
        Callback for auto-saving.
        """
        self.save_to_location()
        return True

    def set_name(self, name: str) -> None:
        """
        Sets the name of this :class:`TrackDB`

        :param name: The new name.
        """
        self.name = name
        self._dirty = True

    def get_name(self) -> str:
        """
        Gets the name of this :class:`TrackDB`

        :return: The name.
        """
        return self.name

    def set_location(self, location: Optional[str]) -> None:
        """
        Sets the location to save to

        :param location: the location to save to
        """
        self.location = location
        self._dirty = True

    @common.synchronized
    def load_from_location(self, location: Optional[str] = None):
        """
        Restores :class:`TrackDB` state from the pickled representation
        stored at the specified location.

        :param location: the location to load the data from
        """
        if not location:
            location = self.location
        if not location:
            raise AttributeError(
                _("You did not specify a location to load the db from")
            )

        logger.debug("Loading %s DB from %s.", self.name, location)

        pdata = common.open_shelf(location)

        if "_dbversion" in pdata:
            if int(pdata['_dbversion']) > int(self._dbversion):
                raise common.VersionError("DB was created on a newer Exaile version.")
            elif pdata['_dbversion'] < self._dbversion:
                logger.info("Upgrading DB format....")
                import shutil

                shutil.copyfile(location, location + "-%s.bak" % pdata['_dbversion'])
                import xl.migrations.database as dbmig

                dbmig.handle_migration(
                    self, pdata, pdata['_dbversion'], self._dbversion
                )

        for attr in self.pickle_attrs:
            try:
                if 'tracks' == attr:
                    data = {}
                    for k in (x for x in pdata.keys() if x.startswith("tracks-")):
                        p = pdata[k]
                        tr = Track(_unpickles=p[0])
                        loc = tr.get_loc_for_io()
                        if loc not in data:
                            data[loc] = TrackHolder(tr, p[1], **p[2])
                        else:
                            logger.warning("Duplicate track found: %s", loc)
                            # presumably the second track was written because of an error,
                            # so use the first track found.
                            del pdata[k]

                    setattr(self, attr, data)
                else:
                    setattr(self, attr, pdata.get(attr, getattr(self, attr)))
            except Exception:
                # FIXME: Do something about this
                logger.exception("Exception occurred while loading %s", location)

        pdata.close()

        self._dirty = False

    @common.synchronized
    def save_to_location(self, location: Optional[str] = None):
        """
        Saves a pickled representation of this :class:`TrackDB` to the
        specified location.

        :param location: the location to save the data to
        """
        if not self._dirty:
            for track in self.tracks.values():
                if track._track._dirty:
                    self._dirty = True
                    break

        if not self._dirty:
            return

        if not location:
            location = self.location
        if not location:
            raise AttributeError(_("You did not specify a location to save the db"))

        if self._saving:
            return
        self._saving = True

        logger.debug("Saving %s DB to %s.", self.name, location)

        try:
            pdata = common.open_shelf(location)
            if pdata.get('_dbversion', self._dbversion) > self._dbversion:
                raise common.VersionError("DB was created on a newer Exaile.")
        except Exception:
            logger.exception("Failed to open music DB for writing.")
            return

        for attr in self.pickle_attrs:
            # bad hack to allow saving of lists/dicts of Tracks
            if 'tracks' == attr:
                for k, track in self.tracks.items():
                    key = "tracks-%s" % track._key
                    if track._track._dirty or key not in pdata:
                        pdata[key] = (
                            track._track._pickles(),
                            track._key,
                            deepcopy(track._attrs),
                        )
            else:
                pdata[attr] = deepcopy(getattr(self, attr))

        pdata['_dbversion'] = self._dbversion

        for key in self._deleted_keys:
            key = "tracks-%s" % key
            if key in pdata:
                del pdata[key]

        pdata.sync()
        pdata.close()

        for track in self.tracks.values():
            track._track._dirty = False

        self._dirty = False
        self._saving = False

    def get_track_by_loc(self, loc: str) -> Optional[Track]:
        """
        returns the track having the given loc. if no such track exists,
        returns None
        """
        try:
            return self.tracks[loc]._track
        except KeyError:
            return None

    def loc_is_member(self, loc: str) -> bool:
        """
        Returns True if loc is a track in this collection, False
        if it is not
        """
        return loc in self.tracks

    def get_count(self) -> int:
        """
        Returns the number of tracks stored in this database
        """
        return len(self.tracks)

    def add(self, track: Track) -> None:
        """
        Adds a track to the database of tracks

        :param track: The :class:`xl.trax.Track` to add
        """
        self.add_tracks([track])

    @common.synchronized
    def add_tracks(self, tracks: Iterable[Track]) -> None:
        """
        Like add(), but takes a list of :class:`xl.trax.Track`
        """
        locations = []
        now = time()
        for tr in tracks:
            if not tr.get_tag_raw('__date_added'):
                tr.set_tags(__date_added=now)
            location = tr.get_loc_for_io()
            # Don't add duplicates -- track URLs are unique
            if location in self.tracks:
                continue
            # Don't add unsupported media files
            if not tr.is_supported():
                continue
            locations += [location]
            self.tracks[location] = TrackHolder(tr, self._key)
            self._key += 1

        if locations:
            event.log_event('tracks_added', self, locations)
            self._dirty = True

    def remove(self, track: Track) -> None:
        """
        Removes a track from the database

        :param track: the :class:`xl.trax.Track` to remove
        """
        self.remove_tracks([track])

    @common.synchronized
    def remove_tracks(self, tracks: Iterable[Track]) -> None:
        """
        Like remove(), but takes a list of :class:`xl.trax.Track`
        """
        locations = []

        for tr in tracks:
            location = tr.get_loc_for_io()
            locations += [location]
            self._deleted_keys.append(self.tracks[location]._key)
            del self.tracks[location]

        event.log_event('tracks_removed', self, locations)

        self._dirty = True

    def get_tracks(self) -> List[Track]:
        return list(self)



import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _("Shutdown")
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "shutdown_preferences.ui")


class ActivateClosingByDefault(widgets.CheckPreference):
    default = False
    name = "shutdown/activate_closing_by_default"


class Timeout(widgets.SpinPreference):
    default = 10
    name = "shutdown/timeout"



# Modified to work with Exaile - Brian Parma
#
# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <https://www.gnu.org/licenses/>.

import http.server
import socketserver
import logging
import select
import socket
import os

import spydaap
import spydaap.daap
import spydaap.metadata
import spydaap.containers
import spydaap.cache
import spydaap.server
import spydaap.zeroconfimpl

from xl import common, event, xdg

# Notes for debugging:
# You might want to run
#    handle SIGPIPE nostop
# when debugging this code in gdb.

"""
Notes for hunting down errors:
If you run this plugin and a client stops playback on any file, expect this
traceback:

    Exception happened during processing of request from ('192.168.122.1', 34394)
    Traceback (most recent call last):
      File "/usr/lib64/python2.7/SocketServer.py", line 596, in process_request_thread
        self.finish_request(request, client_address)
      File "/usr/lib64/python2.7/SocketServer.py", line 331, in finish_request
        self.RequestHandlerClass(request, client_address, self)
      File "/usr/lib64/python2.7/SocketServer.py", line 654, in __init__
        self.finish()
      File "/usr/lib64/python2.7/SocketServer.py", line 713, in finish
        self.wfile.close()
      File "/usr/lib64/python2.7/socket.py", line 283, in close
        self.flush()
      File "/usr/lib64/python2.7/socket.py", line 307, in flush
        self._sock.sendall(view[write_offset:write_offset+buffer_size])
    error: [Errno 32] broken pipe

This traceback is a result of getting a SIGPIPE, which is expected. The fact
that it is not being handled in the python standard library is a bug which has
been reported to https://bugs.python.org/issue14574
See also: https://stackoverflow.com/questions/6063416/
"""


logger = logging.getLogger('daapserver')

__all__ = ['DaapServer']


class MyThreadedHTTPServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    """Handle requests in a separate thread."""

    timeout = 1
    daemon_threads = True

    def __init__(self, *args):
        if ':' in args[0][0]:
            self.address_family = socket.AF_INET6
        http.server.HTTPServer.__init__(self, *args)


class DaapServer:
    def __init__(self, library, name=spydaap.server_name, host='', port=spydaap.port):
        #        Thread.__init__(self)
        self.host = host
        self.port = port
        self.library = library
        self.name = name
        self.httpd = None
        self.handler = None
        self.__cache = spydaap.cache.Cache(os.path.join(xdg.cache_home, 'daapserver'))
        self.__cache.clean()

        # Set a callback that will let us propagate library changes to clients
        event.add_callback(self.update_rev, 'libraries_modified', library.collection)

    def update_rev(self, *args):
        if self.handler is not None:
            # Updating the server revision, so if a client checks
            # it can see the library has changed
            self.handler.daap_server_revision += 1
            logger.info(
                'Libraries Changed, incrementing revision to %d.'
                % self.handler.daap_server_revision
            )
        self.__cache.clean()

    def set(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

    @common.threaded
    def run(self):
        self.zeroconf = spydaap.zeroconfimpl.ZeroconfImpl(
            self.name, self.port, stype="_daap._tcp"
        )
        self.handler = spydaap.server.makeDAAPHandlerClass(
            str(self.name), self.__cache, self.library, []
        )
        self.httpd = MyThreadedHTTPServer((self.host, self.port), self.handler)

        # signal.signal(signal.SIGTERM, make_shutdown(httpd))
        # signal.signal(signal.SIGHUP, rebuild_cache)
        if self.httpd.address_family == socket.AF_INET:
            self.zeroconf.publish(ipv4=True, ipv6=False)
        else:
            self.zeroconf.publish(ipv4=False, ipv6=True)

        try:
            try:
                logger.info("DAAP server: Listening.")
                self.httpd.serve_forever()
            except select.error:
                pass
        except KeyboardInterrupt:
            self.httpd.shutdown()

        logger.info("DAAP server: Shutting down.")
        self.zeroconf.unpublish()
        self.httpd = None

    def start(self):
        if self.httpd is None:
            self.run()
            return True
        return False

    def stop(self):
        if self.httpd is not None:
            self.httpd.shutdown()
            self.httpd.socket.close()
            return True
        return False

    def stop_server(self):
        self.stop()


# def rebuild_cache(signum=None, frame=None):
#    md_cache.build(os.path.abspath(spydaap.media_path))
#    container_cache.clean()
#    container_cache.build(md_cache)
#    cache.clean()

# def really_main():
#    rebuild_cache()
#    zeroconf = spydaap.zeroconfimpl.ZeroconfImpl(spydaap.server_name,
#                                         spydaap.port,
#                                         stype="_daap._tcp")
#    zeroconf.publish()
#    logger.warning("Listening.")
#    httpd = MyThreadedHTTPServer(('0.0.0.0', spydaap.port),
#                                 spydaap.server.makeDAAPHandlerClass(spydaap.server_name, cache, md_cache, container_cache))
#
##    signal.signal(signal.SIGTERM, make_shutdown(httpd))
##    signal.signal(signal.SIGHUP, rebuild_cache)
#
#    try:
#        try:
#            httpd.serve_forever()
#        except select.error:
#            pass
#    except KeyboardInterrupt:
#        httpd.force_stop()
#    logger.warning("Shutting down.")
#    zeroconf.unpublish()

# def main():
#    really_main()

# if __name__ == "__main__":
#    main()



# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk

import os.path
import webbrowser

from xl.nls import gettext as _
from xl import common, settings, providers, xdg

from xlgui.accelerators import Accelerator
from xlgui.widgets import menu, dialogs


def get_main():
    from xlgui import main

    return main.mainwindow()


def get_selected_playlist():
    from xlgui import main

    return main.get_selected_playlist()


_smi = menu.simple_menu_item
_sep = menu.simple_separator


def __create_file_menu():
    items = []
    accelerators = []

    def new_playlist_cb(*args):
        get_main().playlist_container.create_new_playlist()

    accelerators.append(Accelerator('<Primary>t', _("_New Playlist"), new_playlist_cb))
    items.append(
        _smi('new-playlist', [], icon_name='tab-new', callback=accelerators[-1])
    )
    items.append(_sep('new-sep', [items[-1].name]))

    def open_cb(*args):
        dialog = dialogs.MediaOpenDialog(get_main().window)
        dialog.connect(
            'uris-selected', lambda d, uris: get_main().controller.open_uris(uris)
        )
        dialog.show()

    accelerators.append(Accelerator('<Primary>o', _("_Open"), open_cb))
    items.append(
        _smi(
            'open',
            [items[-1].name],
            icon_name='document-open',
            callback=accelerators[-1],
        )
    )

    def open_uri_cb(*args):
        dialog = dialogs.URIOpenDialog(get_main().window)
        dialog.connect(
            'uri-selected', lambda d, uri: get_main().controller.open_uri(uri)
        )
        dialog.show()

    accelerators.append(Accelerator('<Primary><Shift>o', _("Open _URL"), open_uri_cb))
    items.append(
        _smi(
            'open-uri',
            [items[-1].name],
            icon_name='emblem-web',
            callback=accelerators[-1],
        )
    )

    def open_dirs_cb(*args):
        dialog = dialogs.DirectoryOpenDialog(get_main().window)
        dialog.props.create_folders = False
        dialog.connect(
            'uris-selected', lambda d, uris: get_main().controller.open_uris(uris)
        )
        dialog.show()

    items.append(
        _smi(
            'open-dirs',
            [items[-1].name],
            _("Open _Directories"),
            'folder-open',
            open_dirs_cb,
        )
    )

    items.append(_sep('open-sep', [items[-1].name]))

    items.append(
        _smi(
            'import-playlist',
            [items[-1].name],
            _("_Import Playlist"),
            'document-open',
            lambda *e: get_main().controller.get_panel('playlists').import_playlist(),
        )
    )

    def export_playlist_cb(*args):
        main = get_main()
        page = get_selected_playlist()
        if not page:
            return

        def on_message(dialog, message_type, message):
            """
            Show messages in the main window message area
            """
            if message_type == Gtk.MessageType.INFO:
                main.message.show_info(markup=message)
            elif message_type == Gtk.MessageType.ERROR:
                main.message.show_error(_('Playlist export failed!'), message)
            return True

        dialog = dialogs.PlaylistExportDialog(page.playlist, main.window)
        dialog.connect('message', on_message)
        dialog.show()

    items.append(
        _smi(
            'export-playlist',
            [items[-1].name],
            _("E_xport Current Playlist"),
            'document-save-as',
            export_playlist_cb,
        )
    )
    items.append(_sep('export-sep', [items[-1].name]))

    def close_tab_cb(*args):
        get_main().get_selected_page().tab.close()

    accelerators.append(Accelerator('<Primary>w', _("Close _Tab"), close_tab_cb))
    items.append(
        _smi(
            'close-tab',
            [items[-1].name],
            icon_name='window-close',
            callback=accelerators[-1],
        )
    )

    if get_main().controller.exaile.options.Debug:

        def restart_cb(*args):
            from xl import main

            main.exaile().quit(True)

        accelerators.append(Accelerator('<Primary>r', _("_Restart"), restart_cb))
        items.append(
            _smi('restart-application', [items[-1].name], callback=accelerators[-1])
        )

    def quit_cb(*args):
        from xl import main

        main.exaile().quit()

    accelerators.append(Accelerator('<Primary>q', _("_Quit Exaile"), quit_cb))
    items.append(
        _smi(
            'quit-application',
            [items[-1].name],
            icon_name='application-exit',
            callback=accelerators[-1],
        )
    )

    for item in items:
        providers.register('menubar-file-menu', item)
    for accelerator in accelerators:
        providers.register('mainwindow-accelerators', accelerator)


def __create_edit_menu():
    items = []
    accelerators = []

    def collection_manager_cb(*args):
        from xlgui import get_controller

        get_controller().collection_manager()

    items.append(
        _smi('collection-manager', [], _("_Collection"), None, collection_manager_cb)
    )

    def queue_cb(*args):
        get_main().playlist_container.show_queue()

    accelerators.append(Accelerator('<Primary>m', _("_Queue"), queue_cb))
    items.append(_smi('queue', [items[-1].name], callback=accelerators[-1]))

    def cover_manager_cb(*args):
        from xlgui.cover import CoverManager

        CoverManager(get_main().window, get_main().collection)

    items.append(
        _smi(
            'cover-manager',
            [items[-1].name],
            _("C_overs"),
            'image-x-generic',
            cover_manager_cb,
        )
    )

    def preferences_cb(*args):
        from xlgui.preferences import PreferencesDialog

        dialog = PreferencesDialog(get_main().window, get_main().controller)
        dialog.run()

    items.append(
        _smi(
            'preferences',
            [items[-1].name],
            _("_Preferences"),
            'preferences-system',
            preferences_cb,
        )
    )

    for item in items:
        providers.register('menubar-edit-menu', item)
    for accelerator in accelerators:
        providers.register('mainwindow-accelerators', accelerator)


def __create_view_menu():
    items = []
    accelerators = []

    def show_playing_track_cb(*args):
        get_main().playlist_container.show_current_track()

    def show_playing_track_sensitive():
        from xl import player

        return player.PLAYER.get_state() != 'stopped'

    accelerators.append(
        Accelerator('<Primary>j', _("_Show Playing Track"), show_playing_track_cb)
    )
    items.append(
        _smi(
            'show-playing-track',
            [],
            icon_name='go-jump',
            callback=accelerators[-1],
            sensitive_cb=show_playing_track_sensitive,
        )
    )

    items.append(_sep('show-playing-track-sep', [items[-1].name]))

    def playlist_utilities_cb(widget, name, parent, context):
        settings.set_option('gui/playlist_utilities_bar_visible', widget.get_active())

    def playlist_utilities_is_checked(name, parent, context):
        return settings.get_option('gui/playlist_utilities_bar_visible', True)

    items.append(
        menu.check_menu_item(
            'playlist-utilities',
            [items[-1].name],
            _("_Playlist Utilities Bar"),
            playlist_utilities_is_checked,
            playlist_utilities_cb,
        )
    )

    items.append(
        _smi(
            'columns',
            [items[-1].name],
            _('C_olumns'),
            submenu=menu.ProviderMenu('playlist-columns-menu', get_main()),
        )
    )

    def clear_playlist_cb(*args):
        page = get_main().get_selected_page()
        if page:
            page.playlist.clear()

    accelerators.append(
        Accelerator('<Primary>l', _('_Clear playlist'), clear_playlist_cb)
    )
    items.append(
        _smi(
            'clear-playlist',
            [items[-1].name],
            icon_name='edit-clear-all',
            callback=accelerators[-1],
        )
    )

    for item in items:
        providers.register('menubar-view-menu', item)
    for accelerator in accelerators:
        providers.register('mainwindow-accelerators', accelerator)


def __create_playlist_menu():
    items = []
    for item in items:
        providers.register('menubar-playlist-menu', item)


def __create_tools_menu():
    items = []
    items.append(
        _smi(
            'device-manager',
            [],
            _('_Device Manager'),
            'multimedia-player',
            lambda *x: get_main().controller.show_devices(),
        )
    )

    items.append(
        _smi(
            'scan-collection',
            [items[-1].name],
            _('_Rescan Collection'),
            'view-refresh',
            get_main().controller.on_rescan_collection,
        )
    )

    items.append(
        _smi(
            'slow-scan-collection',
            [items[-1].name],
            _('Rescan Collection (_slow)'),
            'view-refresh',
            get_main().controller.on_rescan_collection_forced,
        )
    )

    for item in items:
        providers.register('menubar-tools-menu', item)


def __create_help_menu():
    items = []
    accelerators = []

    def show_logs_directory(*args):
        common.open_file_directory(os.path.join(xdg.get_logs_dir(), 'exaile.log'))

    def show_report_issue(*args):
        webbrowser.open('https://github.com/exaile/exaile/issues')

    def show_shortcuts(widget, name, parent, context):
        dialog = dialogs.ShortcutsDialog(parent.window)
        dialog.show()

    def show_user_guide(*args):
        # TODO: Other languages
        webbrowser.open('https://exaile.readthedocs.io/en/latest/user/index.html')

    def show_about_dialog(widget, name, parent, context):
        dialog = dialogs.AboutDialog(parent.window)
        dialog.show()

    items.append(
        _smi('guide', [], _("User's Guide (website)"), 'help-contents', show_user_guide)
    )
    items.append(
        _smi('shortcuts', [items[-1].name], _("Shortcuts"), None, show_shortcuts)
    )
    items.append(_sep('about-sep1', [items[-1].name]))
    items.append(
        _smi(
            'report',
            [items[-1].name],
            _("Report an issue (GitHub)"),
            None,
            show_report_issue,
        )
    )
    items.append(
        _smi('logs', [items[-1].name], _("Open error logs"), None, show_logs_directory)
    )
    items.append(_sep('about-sep2', [items[-1].name]))
    items.append(
        _smi('about', [items[-1].name], _("_About"), 'help-about', show_about_dialog)
    )
    for item in items:
        providers.register('menubar-help-menu', item)
    for accelerator in accelerators:
        providers.register('mainwindow-accelerators', accelerator)


def _create_menus():
    __create_file_menu()
    __create_edit_menu()
    __create_view_menu()
    __create_playlist_menu()
    __create_tools_menu()
    __create_help_menu()



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import locale
import logging
import os

from gi.repository import Gdk, GdkPixbuf, Gio, GLib, GObject, Gtk, Pango

from xl import common, event, metadata, settings, trax
from xl.nls import gettext as _
from xl.trax.util import recursive_tracks_from_file
from xlgui import guiutil, icons, panel, xdg

from xlgui.panel import menus
from xlgui.widgets.common import DragTreeView

logger = logging.getLogger(__name__)


def gfile_enumerate_children(gfile, attributes, follow_symlinks=True):
    """Like Gio.File.enumerate_children but ignores errors"""
    flags = (
        Gio.FileQueryInfoFlags.NONE
        if follow_symlinks
        else Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS
    )
    infos = gfile.enumerate_children(attributes, flags, None)
    it = iter(infos)
    while True:
        try:
            yield next(it)
        except StopIteration:
            break
        except GLib.Error:
            logger.warning(
                "Error while iterating on %r", gfile.get_parse_name(), exc_info=True
            )


class FilesPanel(panel.Panel):
    """
    The Files panel
    """

    __gsignals__ = {
        'append-items': (GObject.SignalFlags.RUN_LAST, None, (object, bool)),
        'replace-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'queue-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    ui_info = ('files.ui', 'FilesPanel')

    def __init__(self, parent, collection, name):
        """
        Initializes the files panel
        """
        panel.Panel.__init__(self, parent, name, _('Files'))
        self.collection = collection

        self.box = self.builder.get_object('FilesPanel')

        self.targets = [Gtk.TargetEntry.new('text/uri-list', 0, 0)]

        self._setup_tree()
        self._setup_widgets()
        self.menu = menus.FilesContextMenu(self)

        self.key_id = None
        self.i = 0

        first_dir = Gio.File.new_for_commandline_arg(
            settings.get_option('gui/files_panel_dir', xdg.homedir)
        )
        self.history = [first_dir]
        self.load_directory(first_dir, False)

    def _setup_tree(self):
        """
        Sets up tree widget for the files panel
        """
        self.model = Gtk.ListStore(Gio.File, GdkPixbuf.Pixbuf, str, str, bool)
        self.tree = tree = FilesDragTreeView(self, receive=False, source=True)
        tree.set_model(self.model)
        tree.connect('row-activated', self.row_activated)
        tree.connect('key-release-event', self.on_key_released)

        selection = tree.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.scroll = scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.add(tree)
        scroll.set_shadow_type(Gtk.ShadowType.IN)
        self.box.pack_start(scroll, True, True, 0)

        pb = Gtk.CellRendererPixbuf()
        text = Gtk.CellRendererText()
        self.colname = colname = Gtk.TreeViewColumn(_('Filename'))
        colname.pack_start(pb, False)
        colname.pack_start(text, True)
        if settings.get_option('gui/ellipsize_text_in_panels', False):
            text.set_property('ellipsize-set', True)
            text.set_property('ellipsize', Pango.EllipsizeMode.END)
        else:
            colname.connect('notify::width', self.set_column_width)

            width = settings.get_option('gui/files_filename_col_width', 130)

            colname.set_fixed_width(width)
            colname.set_sizing(Gtk.TreeViewColumnSizing.FIXED)

        colname.set_resizable(True)
        colname.set_attributes(pb, pixbuf=1)
        colname.set_attributes(text, text=2)
        colname.set_expand(True)

        tree.append_column(self.colname)

        text = Gtk.CellRendererText()
        text.set_property('xalign', 1.0)
        # TRANSLATORS: File size column in the file browser
        self.colsize = colsize = Gtk.TreeViewColumn(_('Size'))
        colsize.set_resizable(True)
        colsize.pack_start(text, False)
        colsize.set_attributes(text, text=3)
        colsize.set_expand(False)
        tree.append_column(colsize)

    def _setup_widgets(self):
        """
        Sets up the widgets for the files panel
        """
        self.directory = icons.MANAGER.pixbuf_from_icon_name(
            'folder', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.track = icons.MANAGER.pixbuf_from_icon_name(
            'audio-x-generic', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.back = self.builder.get_object('files_back_button')
        self.back.connect('clicked', self.go_back)
        self.forward = self.builder.get_object('files_forward_button')
        self.forward.connect('clicked', self.go_forward)
        self.up = self.builder.get_object('files_up_button')
        self.up.connect('clicked', self.go_up)
        self.builder.get_object('files_refresh_button').connect('clicked', self.refresh)
        self.builder.get_object('files_home_button').connect('clicked', self.go_home)

        # Set up the location bar
        self.location_bar = self.builder.get_object('files_entry')
        self.location_bar.connect('changed', self.on_location_bar_changed)
        event.add_ui_callback(
            self.fill_libraries_location, 'libraries_modified', self.collection
        )
        self.fill_libraries_location()
        self.location_bar.set_row_separator_func(lambda m, i: m[i][1] is None)
        self.entry = self.location_bar.get_children()[0]
        self.entry.connect('activate', self.entry_activate)

        # Set up the search entry
        self.filter = guiutil.SearchEntry(self.builder.get_object('files_search_entry'))
        self.filter.connect(
            'activate',
            lambda *e: self.load_directory(
                self.current,
                history=False,
                keyword=self.filter.get_text(),
            ),
        )

    def fill_libraries_location(self, *e):
        libraries = []
        for library in self.collection._serial_libraries:
            f = Gio.File.new_for_commandline_arg(library['location'])
            libraries.append((f.get_parse_name(), f.get_uri()))

        mounts = []
        for mount in Gio.VolumeMonitor.get().get_mounts():
            name = mount.get_name()
            uri = mount.get_default_location().get_uri()
            mounts.append((name, uri))
        mounts.sort(key=lambda row: locale.strxfrm(row[0]))

        model = self.location_bar.get_model()
        model.clear()
        for row in libraries:
            model.append(row)
        if libraries and mounts:
            model.append((None, None))
        for row in mounts:
            model.append(row)
        self.location_bar.set_model(model)

    def on_location_bar_changed(self, widget, *args):
        # Find out which one is selected, if any.
        iter = self.location_bar.get_active_iter()
        if not iter:
            return
        model = self.location_bar.get_model()
        uri = model.get_value(iter, 1)
        if uri:
            self.load_directory(Gio.File.new_for_uri(uri))

    def on_key_released(self, widget, event):
        """
        Called when a key is released in the tree
        """
        if event.keyval == Gdk.KEY_Menu:
            Gtk.Menu.popup(self.menu, None, None, None, None, 0, event.time)
            return True

        if (
            event.keyval == Gdk.KEY_Left
            and Gdk.ModifierType.MOD1_MASK & event.get_state()
        ):
            self.go_back(self.tree)
            return True

        if (
            event.keyval == Gdk.KEY_Right
            and Gdk.ModifierType.MOD1_MASK & event.get_state()
        ):
            self.go_forward(self.tree)
            return True

        if (
            event.keyval == Gdk.KEY_Up
            and Gdk.ModifierType.MOD1_MASK & event.get_state()
        ):
            self.go_up(self.tree)
            return True

        if event.keyval == Gdk.KEY_BackSpace:
            self.go_up(self.tree)
            return True

        if event.keyval == Gdk.KEY_F5:
            self.refresh(self.tree)
            return True
        return False

    def row_activated(self, *i):
        """
        Called when someone double clicks a row
        """
        selection = self.tree.get_selection()
        model, paths = selection.get_selected_rows()

        for path in paths:
            if model[path][4]:
                self.load_directory(model[path][0])
            else:
                self.emit('append-items', self.tree.get_selected_tracks(), True)

    def refresh(self, widget):
        """
        Refreshes the current view
        """
        treepath = self.tree.get_cursor()[0]
        cursorf = self.model[treepath][0] if treepath else None
        self.load_directory(self.current, history=False, cursor_file=cursorf)
        self.fill_libraries_location()

    def entry_activate(self, widget, event=None):
        """
        Called when the user presses enter in the entry box
        """
        path = self.entry.get_text()
        if path.startswith('~'):
            path = os.path.expanduser(path)
        f = Gio.file_parse_name(path)
        try:
            ftype = f.query_info(
                'standard::type', Gio.FileQueryInfoFlags.NONE, None
            ).get_file_type()
        except GLib.Error as e:
            logger.exception(e)
            self.entry.set_text(self.current.get_parse_name())
            return
        if ftype != Gio.FileType.DIRECTORY:
            f = f.get_parent()
        self.load_directory(f)

    def focus(self):
        self.tree.grab_focus()

    def go_forward(self, widget):
        """
        Goes to the next entry in history
        """
        assert 0 <= self.i < len(self.history)
        if self.i == len(self.history) - 1:
            return
        self.i += 1
        self.load_directory(
            self.history[self.i], history=False, cursor_file=self.current
        )
        if self.i >= len(self.history) - 1:
            self.forward.set_sensitive(False)
        if self.history:
            self.back.set_sensitive(True)

    def go_back(self, widget):
        """
        Goes to the previous entry in history
        """
        assert 0 <= self.i < len(self.history)
        if self.i == 0:
            return
        self.i -= 1
        self.load_directory(
            self.history[self.i], history=False, cursor_file=self.current
        )
        if self.i == 0:
            self.back.set_sensitive(False)
        if self.history:
            self.forward.set_sensitive(True)

    def go_up(self, widget):
        """
        Moves up one directory
        """
        parent = self.current.get_parent()
        if parent:
            self.load_directory(parent, cursor_file=self.current)

    def go_home(self, widget):
        """
        Goes to the user's home directory
        """
        home = Gio.File.new_for_commandline_arg(xdg.homedir)
        if home.get_uri() == self.current.get_uri():
            self.refresh(widget)
        else:
            self.load_directory(home, cursor_file=self.current)

    def set_column_width(self, col, stuff=None):
        """
        Called when the user resizes a column
        """
        name = {self.colname: 'filename', self.colsize: 'size'}[col]
        name = "gui/files_%s_col_width" % name

        # this option gets triggered all the time, which is annoying when debugging,
        # so only set it when it actually changes
        w = col.get_width()
        if settings.get_option(name, w) != w:
            settings.set_option(name, w, save=False)

    @common.threaded
    def load_directory(self, directory, history=True, keyword=None, cursor_file=None):
        """
        Load a directory into the files view.

        :param history: whether to record in history
        :param keyword: filter string
        :param cursor_file: file to (attempt to) put the cursor on.
            Will put the cursor on a subdirectory if the file is under it.
        """
        self.current = directory
        try:
            infos = gfile_enumerate_children(
                directory,
                'standard::display-name,standard::is-hidden,standard::name,standard::type',
            )
        except GLib.Error as e:
            logger.exception(e)
            if directory.get_path() != xdg.homedir:  # Avoid infinite recursion.
                self.load_directory(
                    Gio.File.new_for_commandline_arg(xdg.homedir),
                    history,
                    keyword,
                    cursor_file,
                )
            return
        if self.current != directory:  # Modified from another thread.
            return

        settings.set_option('gui/files_panel_dir', directory.get_uri())

        subdirs = []
        subfiles = []
        for info in infos:
            if info.get_is_hidden():
                # Ignore hidden files. They can still be accessed manually from
                # the location bar.
                continue
            name = info.get_display_name()
            low_name = name.lower()
            if keyword and keyword.lower() not in low_name:
                continue
            f = directory.get_child(info.get_name())

            ftype = info.get_file_type()
            sortname = locale.strxfrm(name)
            if ftype == Gio.FileType.DIRECTORY:
                subdirs.append((sortname, name, f))
            elif any(low_name.endswith('.' + ext) for ext in metadata.formats):
                subfiles.append((sortname, name, f))

        subdirs.sort()
        subfiles.sort()

        def idle():
            if self.current != directory:  # Modified from another thread.
                return

            model = self.model
            view = self.tree

            if cursor_file:
                cursor_uri = cursor_file.get_uri()
            cursor_row = -1

            model.clear()
            row = 0
            for sortname, name, f in subdirs:
                model.append((f, self.directory, name, '', True))
                uri = f.get_uri()
                if (
                    cursor_file
                    and cursor_row == -1
                    and (cursor_uri == uri or cursor_uri.startswith(uri + '/'))
                ):
                    cursor_row = row
                row += 1
            for sortname, name, f in subfiles:
                size = (
                    f.query_info(
                        'standard::size', Gio.FileQueryInfoFlags.NONE, None
                    ).get_size()
                    // 1000
                )

                # TRANSLATORS: File size (1 kB = 1000 bytes)
                size = _('%s kB') % locale.format_string('%d', size, True)

                model.append((f, self.track, name, size, False))
                if cursor_file and cursor_row == -1 and cursor_uri == f.get_uri():
                    cursor_row = row
                row += 1

            if cursor_file and cursor_row != -1:
                view.set_cursor((cursor_row,))
            else:
                view.set_cursor((0,))
                if view.get_realized():
                    view.scroll_to_point(0, 0)

            self.entry.set_text(directory.get_parse_name())
            if history:
                self.back.set_sensitive(True)
                self.history[self.i + 1 :] = [self.current]
                self.i = len(self.history) - 1
                self.forward.set_sensitive(False)
            self.up.set_sensitive(bool(directory.get_parent()))

        GLib.idle_add(idle)

    def drag_get_data(self, treeview, context, selection, target_id, etime):
        """
        Called when a drag source wants data for this drag operation
        """
        tracks = self.tree.get_selected_tracks()
        if not tracks:
            return
        for track in tracks:
            DragTreeView.dragged_data[track.get_loc_for_io()] = track
        uris = trax.util.get_uris_from_tracks(tracks)
        selection.set_uris(uris)


class FilesDragTreeView(DragTreeView):
    """
    Custom DragTreeView to retrieve data from files
    """

    def get_selection_empty(self):
        '''Returns True if there are no selected items'''
        return self.get_selection().count_selected_rows() == 0

    def get_selection_is_computed(self):
        """
        Returns True if anything in the selection is a directory
        """
        selection = self.get_selection()
        model, paths = selection.get_selected_rows()

        for path in paths:
            if model[path][4]:
                return True

        return False

    def get_selected_tracks(self):
        """
        Returns the currently selected tracks
        """
        selection = self.get_selection()
        model, paths = selection.get_selected_rows()
        tracks = []

        for path in paths:
            f = model[path][0]
            self.append_recursive(tracks, f)

        return trax.sort_tracks(common.BASE_SORT_TAGS, tracks, artist_compilations=True)

    def append_recursive(self, songs, f):
        """
        Appends recursively
        """
        ftype = f.query_info(
            'standard::type', Gio.FileQueryInfoFlags.NONE, None
        ).get_file_type()
        if ftype == Gio.FileType.DIRECTORY:
            file_infos = gfile_enumerate_children(f, 'standard::name')
            files = (f.get_child(fi.get_name()) for fi in file_infos)
            for subf in files:
                self.append_recursive(songs, subf)
        else:
            tr = self.get_track(f)
            if tr:
                songs.append(tr)

    def get_track(self, f):
        """
        Returns a single track from a Gio.File
        """
        uri = f.get_uri()
        if not trax.is_valid_track(uri):
            return None
        tr = trax.Track(uri)
        return tr

    def get_tracks_for_path(self, path):
        """
        Get tracks for a path from model (expand item)
        :param path: Gtk.TreePath
        :return: list of tracks [xl.trax.Track]
        """
        return recursive_tracks_from_file(self.get_model()[path][0])



import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _("Quickbuttons")
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "qb_prefs.ui")


class EnqueueByDefault(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_enqueue_by_default"


class DisableNewTrack(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_disable_new_track_when_playing"


class RemoveItem(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_remove_item_when_played"


class AutoAdvance(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_auto_advance"


class AutoAdvanceDelay(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_auto_advance_delay"


class Equalizer(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_equalizer"


class AudioDevice(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_audio_device"


class AudioDevicePreview(widgets.CheckPreference):
    default = True
    name = "quickbuttons/btn_audio_device_preview"



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import os
import sys
from gi.repository import GLib

# exaile_dir is now defined in xl.version, but most of the codebase
# still accessed it via xl.xdg.exaile_dir
from .version import exaile_dir

homedir = os.path.expanduser("~")
lastdir = homedir

data_home = GLib.get_user_data_dir()
data_home = os.path.join(data_home, "exaile")

config_home = GLib.get_user_config_dir()
config_home = os.path.join(config_home, "exaile")

cache_home = GLib.get_user_cache_dir()
cache_home = os.path.join(cache_home, "exaile")

if sys.platform == 'win32':
    logs_home = os.path.join(data_home, "logs")
else:
    logs_home = os.path.join(cache_home, "logs")

data_dirs = os.getenv("XDG_DATA_DIRS")
if data_dirs is None:
    if sys.platform == 'win32':
        data_dirs = [exaile_dir]
    else:
        data_dirs = ["/usr/local/share/exaile", "/usr/share/exaile"]
else:
    data_dirs = [os.path.join(d, "exaile") for d in data_dirs.split(os.pathsep)]


config_dirs = os.getenv("XDG_CONFIG_DIRS")
if config_dirs is None:
    if sys.platform == 'win32':
        config_dirs = [exaile_dir]
    else:
        config_dirs = ["/usr/local/etc/xdg/exaile", "/etc/xdg/exaile"]
else:
    config_dirs = [os.path.join(d, "exaile") for d in config_dirs.split(os.pathsep)]

fhs_compliant = True
# Detect if Exaile is not installed.
if os.path.exists(os.path.join(exaile_dir, 'data')):
    fhs_compliant = False
    # Insert the "data" directory to data_dirs.
    data_dir = os.path.join(exaile_dir, 'data')
    data_dirs.insert(0, data_dir)
    # insert the config dir
    config_dir = os.path.join(exaile_dir, 'data', 'config')
    config_dirs.insert(0, config_dir)

data_dirs.insert(0, data_home)


def get_config_dir():
    return config_home


def get_config_dirs():
    return config_dirs[:]


def get_data_dir():
    return data_home


def get_data_dirs():
    return data_dirs[:]


def get_cache_dir():
    return cache_home


def get_logs_dir():
    return logs_home


def _get_path(basedirs, *subpath_elements, **kwargs):
    check_exists = kwargs.get("check_exists", True)
    subpath = os.path.join(*subpath_elements)
    for d in basedirs:
        path = os.path.join(d, subpath)
        if not check_exists or os.path.exists(path):
            return path
    return None


def get_data_path(*subpath_elements, **kwargs):
    return _get_path(data_dirs, *subpath_elements, **kwargs)


def get_config_path(*subpath_elements, **kwargs):
    return _get_path(config_dirs, *subpath_elements, **kwargs)


def get_data_home_path(*subpath_elements, **kwargs):
    return _get_path([data_home], *subpath_elements, **kwargs)


def get_last_dir():
    return lastdir


def get_plugin_data_dir():
    path = os.path.join(get_data_dirs()[0], 'plugin_data')
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def _make_missing_dirs():
    """
    Make any missing base XDG directories

    called by the main Exaile object, should not be used elsewhere.
    """
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    if not os.path.exists(config_home):
        os.makedirs(config_home)
    if not os.path.exists(cache_home):
        os.makedirs(cache_home)
    if not os.path.exists(logs_home):
        os.makedirs(logs_home)


plugin_dirs = [os.path.join(p, 'plugins') for p in get_data_dirs()]
if not fhs_compliant:
    plugin_dirs.insert(1, os.path.join(exaile_dir, 'plugins'))

try:
    os.makedirs(plugin_dirs[0])
except Exception:
    pass

plugin_dirs = [x for x in plugin_dirs if os.path.exists(x)]


def get_user_plugin_dir():
    return plugin_dirs[0]


def get_plugin_dirs():
    return plugin_dirs


# vim: et sts=4 sw=4
